/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "context.h"
#include "global.h"
#include "strings.h"
#include "files.h"
#include "lang.h"

#include <lang/context.h>
#include <lang/block.h>
#include <lang/node.h>
#include <lang/object.h>
#include <lang/function.h>
#include <lang/class.h>
#include <lang/var_dep.h>
#include <lang/builtin.h>
#include <lang/compute.h>
#include <common/mempool_core.h>
#include <common/debug.h>
#include <common/file.h>
#include <common/config.h>

#include <stddef.h>

struct cache_context
{
  size_t var_count;
  size_t parent_count;
  unsigned filename;
  unsigned token_number;
  struct lang_node *class_code;
  size_t dep_list;
};

struct context_list
{
  struct context *head, *tail;
  size_t size;
};

#define LIST_PREFIX(X) context_list_##X
#define LIST_CONTAINER struct context_list
#define LIST_NODE struct context
#define LIST_HEAD head
#define LIST_TAIL tail
#define LIST_NEXT cache_next
#define LIST_PREV cache_prev
#define LIST_SIZE_VAL size
#include <common/list.h>

static struct context_list output, input;

struct overwrite_node
{
  struct context *context;
  struct overwrite_node *next;
};

#define LIST_PREFIX(X) overwrite_list_##X
#define LIST_NODE struct overwrite_node
#define LIST_CONTAINER struct overwrite_node
#define LIST_HEAD next
#define LIST_NEXT next
#include <common/list.h>

#define TRIE_PREFIX(X) overwrite_trie_##X
#define TRIE_DATA struct overwrite_node *
#define TRIE_DATA_DEFAULT NULL
#include <common/trie.h>

static struct overwrite_trie_t overwrite_trie;

static void overwrite_insert(struct context *context, bool overwrite, struct mempool_core *tmp_pool)
{
  if(!context->object)
    return;
  struct variable *target = variable_compute(context, "TARGET", tmp_pool);
  size_t max = 0;
  for(size_t i = 0; i < target->count; i ++)
  {
    assert(target->values[i].string);
    if(max < strlen(target->values[i].string))
      max = strlen(target->values[i].string);
  }
  char *end, *id = object_id_get(context, tmp_pool, &end, max + 1);
  size_t offset = end - id;
  for(size_t i = 0; i < target->count; i ++)
  {
    // Prepare the ID
    strcpy(end, target->values[i].string);
    // Find the node
    struct overwrite_node *list = overwrite_trie_find(&overwrite_trie, id);
    if(list)
    {
      for(struct overwrite_node *node = list; node; node = node->next)
      {
        if(node->context == context)
          goto NEXT;
        if(!strcmp(node->context->filename, context->filename) && (node->context->token == context->token))
        {
          assert(overwrite);
          if(!-- node->context->cache_refs)
          {
            dprint("Context %s/%u is deprecated", node->context->filename, node->context->token);
            context_list_remove(&input, node->context);
            context_return(node->context);
          }
          node->context = context;
          goto NEXT;
        }
      }
      struct overwrite_node *node = mempool_alloc(&persistent_pool, sizeof *node);
      node->context = context;
      overwrite_list_prepend(list, node);
    } else {
      list = mempool_alloc(&persistent_pool, sizeof *list);
      list->context = context;
      overwrite_trie_insert(&overwrite_trie, id, list);
    }
    NEXT:;
  }
}

struct cache_var
{
  unsigned id;  // So the same variables in multiple context can be identified
  unsigned name;
  size_t value_count; // Makes sense only if cache_id is one bigger than previous valid cache_id
                      // Otherwise find the already loaded var
  size_t dep_list;
};

struct cache_var_value
{
  enum value_type type:3;
  unsigned builtin:10;    // For builtin functions, classes, etc
  unsigned string;
  unsigned context, func_context;
  struct lang_node *expr, *object_class;
};

struct context_fill
{
  unsigned cache_var_id;
  size_t context_num;
  size_t var_count, var_count_total;
  size_t var_value_count;
  size_t parent_count;
  struct cache_context *contexts;
  struct context *context;
};

static size_t context_count, var_count, var_value_count, parent_count_in;
static struct cache_context *contexts_in;
static struct cache_var *vars_in;
static struct cache_var_value *var_values_in;
static size_t dep_list_count_in, *dep_list_counts_in;
static unsigned *var_deps_in;
static struct var_dep_list *dep_lists_in;
static struct context **contexts_in_real;
static size_t *parents_in;

static void prepare_var(struct var_trie_node *node, const char *name, struct context_fill *data)
{
  struct var_holder *holder = node->data;
  // Only new variables make it trough
  if(holder->variable->state == vs_cache)
  {
    struct variable *new_var = cache_var_check(holder->variable, data->context == context_toplevel);
    if(new_var)
      holder->variable = new_var;
  }
  // Only computed variables. The rest is not interesting to store
  if(holder->variable->state != vs_computed || holder->variable->undefined)
    return;
  // Another variable
  data->var_count ++;
  // Not yet stored, give it an id and count variable values
  if(!holder->variable->cache_id)
  {
    holder->variable->cache_id = ++ data->cache_var_id;
    data->var_value_count += holder->variable->count;
  }
}

static void parent_count(struct context_plug *plug, size_t *count)
{
  if(plug->context->builtin)
    return;
  assert(plug->context->cache_scheduled);
  (*count) ++;
}

static void context_fill(struct context *context, struct context_fill *data)
{
  assert(context->cache_scheduled);
  // Scan variables
  data->var_count = 0;
  data->context = context;
  var_trie_walk(&context->variables, data, (var_trie_walk_func) prepare_var);
  // Put there the context
  size_t parent_count_tmp = 0;
  plug_list_walk(&context->parents, &parent_count_tmp, (plug_list_walk_func) parent_count);
  data->parent_count += parent_count_tmp;
  *data->contexts ++ = (struct cache_context) {
    .var_count = data->var_count,
    .filename = cache_string_push(context->filename),
    .token_number = context->token,
    .class_code = context->object ? context->object->class_code : NULL,
    .parent_count = parent_count_tmp,
    .dep_list = context->dep_list ? context->dep_list->cache_number + 1 : 0
  };
  context->context_num = ++ data->context_num;
  data->var_count_total += data->var_count;
}

struct var_fill
{
  struct cache_var *vars;
  struct cache_var_value *var_values;
  unsigned last_var_id;
  size_t *parents;
};

static void var_value_store(struct cache_var_value *result, struct var_value *value)
{
  result->type = value->type;
  switch(value->type)
  {
    case vt_class:
    case vt_function:
    case vt_method:
      result->builtin = 0;
      if(value->function->code != ((value->type == vt_class) ? object_build : function_user_eval))
        while(builtin_functions[result->builtin ++].code != value->function->code);
      assert(!value->function->func_context || value->function->func_context->cache_scheduled);
      result->expr = value->function->func;
      result->func_context = value->function->func_context ? value->function->func_context->context_num : 0;
      if(value->type != vt_method)
        break;
      // Method falls trough and fills in the object
    case vt_object:
      assert(value->object->context->cache_next || value->object->context == output.tail);
      assert(value->object->context->cache_scheduled);
      result->context = value->object->context->context_num;
      result->object_class = value->object->class_code;
      // Fall trough and fill in the name of object
    case vt_string:
      result->string = cache_string_push(value->string);
      break;
  }
}

static void var_fill_internal(struct var_trie_node *node, const char *name, struct var_fill *data)
{
  struct variable *var = node->data->variable;
  if(var->state != vs_computed || var->undefined)
    return;
  *data->vars ++ = (struct cache_var) {
    .id = var->cache_id,
    .name = cache_string_push_dup(name),
    .value_count = var->count,
    .dep_list = var->deps ? var->deps->cache_number + 1 : 0
  };
  // Only if not already stored
  if(var->cache_id > data->last_var_id)
  {
    data->last_var_id ++;
    for(size_t i = 0; i < var->count; i ++)
      var_value_store(data->var_values ++, &var->values[i]);
  }
}

static void parent_fill(struct context_plug *plug, size_t **result)
{
  if(plug->context->builtin)
    return;
  *(*result) ++ = plug->context->context_num;
  dprint("Context-Parent: %u %s:%u", plug->context->context_num, plug->context->filename, plug->context->token);
}

static void var_fill(struct context *context, struct var_fill *data)
{
  dprint("Context-Child: %u %s:%u", context->context_num, context->filename, context->token);
  var_trie_walk(&context->variables, data, (var_trie_walk_func) var_fill_internal);
  plug_list_walk(&context->parents, &data->parents, (plug_list_walk_func) parent_fill);
}

static void deps_number(struct var_dep_list *list, size_t *id)
{
  list->cache_number = (*id) ++;
}

struct deps_scan
{
  size_t total;
  size_t *next;
};

static void deps_scan(struct var_dep_list *list, struct deps_scan *data)
{
  data->total += (*data->next ++ = dep_list_size(list));
}

static void chain_fill_single(struct var_dep *dep, unsigned **string_buff)
{
  *(*string_buff) ++ = cache_string_push(dep->filename);
}

static void chain_fill(struct var_dep_list *list, unsigned **string_buff)
{
  dep_list_walk(list, string_buff, (dep_list_walk_func) chain_fill_single);
}

static void context_scan_check(struct context_plug *plug, void *unused)
{
  struct context *context = plug->context;
  if(context->node && !context->object)
  {
    context->object = mempool_alloc(&cache_pool_input, sizeof *context->object);
    *context->object = (struct object) {
      .context = context,
      .class_code = context->node,
      .cache_loaded = true
    };
  }
  if(!context->block)
  {
    if(!context->node) // Static block
      return;
    context->block = block_scan_class(context->node->children_tail);
    context_prepare(context);
    plug_list_walk(&context->parents, NULL, context_scan_check);
  }
}

static void mark_scheduled(struct context *context, void *unused)
{
  context->cache_scheduled = true;
  if(context->node && !context->block)
    context_scan_check(&(struct context_plug) { .context = context }, NULL);
}

void cache_context_store(int fd)
{
  dprinte("Storing contexts to cache");
  // Prepare the variable dependencies
  size_t dep_id = 0;
  dep_chain_walk(&var_dep_single, &dep_id, (dep_chain_walk_func) deps_number);
  dep_chain_walk(&var_dep_merged, &dep_id, (dep_chain_walk_func) deps_number);
  cache_write(fd, &dep_id, sizeof dep_id);
  size_t *dep_buff = mempool_borrow(&cache_pool_output, dep_id * sizeof *dep_buff);
  struct deps_scan dep_scan_data = {
    .next = dep_buff
  };
  dep_chain_walk(&var_dep_single, &dep_scan_data, (dep_chain_walk_func) deps_scan);
  dep_chain_walk(&var_dep_merged, &dep_scan_data, (dep_chain_walk_func) deps_scan);
  cache_write(fd, dep_buff, dep_id * sizeof *dep_buff);
  unsigned *dep_files = mempool_alloc(&cache_pool_output, dep_scan_data.total * sizeof *dep_files),
      *dep_files_pos = dep_files;
  dep_chain_walk(&var_dep_single, &dep_files_pos, (dep_chain_walk_func) chain_fill);
  dep_chain_walk(&var_dep_merged, &dep_files_pos, (dep_chain_walk_func) chain_fill);
  cache_write(fd, dep_files, dep_scan_data.total * sizeof *dep_files);
  // Few cheap magic tricks with toplevel contexts
  context_toplevel->cache_scheduled = true;
  context_list_prepend(&output, context_toplevel);
  if(input.head)
    context_list_remove(&input, input.head);
  // Store the not picked-up contexts too, so they are not lost
  context_list_walk(&input, NULL, mark_scheduled);
  context_list_concat_append(&output, &input);
  // Scan variables and store contexts
  size_t contexts_size;
  struct cache_context *contexts = mempool_alloc(&cache_pool_output, contexts_size = output.size * sizeof *contexts);
  struct context_fill context_fill_data = {
    .contexts = contexts
  };
  context_list_walk(&output, &context_fill_data, (context_list_walk_func) context_fill);
  // Output what you have already
  cache_write(fd, &output.size, sizeof output.size);
  cache_write(fd, &context_fill_data.var_count_total, sizeof context_fill_data.var_count_total);
  cache_write(fd, &context_fill_data.var_value_count, sizeof context_fill_data.var_value_count);
  cache_write(fd, &context_fill_data.parent_count, sizeof context_fill_data.parent_count);
  cache_write(fd, contexts, contexts_size);
  // Buffers for the variables
  size_t vars_size, var_values_size, parents_size;
  struct cache_var *vars = mempool_alloc(&cache_pool_output,
      vars_size = context_fill_data.var_count_total * sizeof *vars);
  struct cache_var_value *var_values = mempool_alloc(&cache_pool_output,
      var_values_size = context_fill_data.var_value_count * sizeof *var_values);
  size_t *parents = mempool_alloc(&cache_pool_output, parents_size = context_fill_data.parent_count * sizeof *parents);
  // Prepare the variables
  struct var_fill var_fill_data = {
    .vars = vars,
    .var_values = var_values,
    .parents = parents
  };
  context_list_walk(&output, &var_fill_data, (context_list_walk_func) var_fill);
  assert(var_fill_data.parents == parents + context_fill_data.parent_count);
  // Output the variables
  cache_write(fd, vars, vars_size);
  cache_write(fd, var_values, var_values_size);
  cache_write(fd, parents, parents_size);
  dprint("Contexts: %zu, vars: %zu, values: %zu, parents: %zu", output.size, context_fill_data.var_count_total,
      context_fill_data.var_value_count, context_fill_data.parent_count);
}

void cache_context_load(int fd)
{
  dprinte("Loading contexts from cache");
  // Deps
  cache_read(fd, &dep_list_count_in, sizeof dep_list_count_in);
  dep_list_counts_in = mempool_alloc(&cache_pool_input_tmp, dep_list_count_in * sizeof *dep_list_counts_in);
  cache_read(fd, dep_list_counts_in, dep_list_count_in * sizeof dep_list_counts_in);
  size_t dep_count = 0;
  for(size_t i = 0; i < dep_list_count_in; i ++)
    dep_count += dep_list_counts_in[i];
  var_deps_in = mempool_alloc(&cache_pool_input_tmp, dep_count * sizeof *var_deps_in);
  cache_read(fd, var_deps_in, dep_count * sizeof *var_deps_in);
  // How many of them?
  cache_read(fd, &context_count, sizeof context_count);
  cache_read(fd, &var_count, sizeof var_count);
  cache_read(fd, &var_value_count, sizeof var_value_count);
  cache_read(fd, &parent_count_in, sizeof parent_count_in);
  // Buffers
  size_t contexts_size, vars_size, var_values_size, parents_size;
  contexts_in = mempool_alloc(&cache_pool_input_tmp, contexts_size = context_count * sizeof *contexts_in);
  vars_in = mempool_alloc(&cache_pool_input_tmp, vars_size = var_count * sizeof *vars_in);
  var_values_in = mempool_alloc(&cache_pool_input, var_values_size = var_value_count * sizeof *var_values_in);
  parents_in = mempool_alloc(&cache_pool_input_tmp, parents_size = parent_count_in * sizeof *parents_in);
  // Load them
  cache_read(fd, contexts_in, contexts_size);
  cache_read(fd, vars_in, vars_size);
  cache_read(fd, var_values_in, var_values_size);
  cache_read(fd, parents_in, parents_size);
  dprint("Contexts: %zu, vars: %zu, values: %zu, parents: %zu", context_count, var_count, var_value_count,
      parent_count_in);
}

static struct context *context_object_resolve(size_t index)
{
  struct context *result = contexts_in_real[index];
  assert(result);
  assert(result->cache_valid);
  context_scan_check(&(struct context_plug) { .context = result }, NULL);
  return result;
}

static void var_value_load(struct var_value *result, struct cache_var_value *src, struct variable *compiled)
{
  result->type = src->type;
  switch(src->type)
  {
    case vt_class:
      assert(contexts_in_real[src->func_context - 1]);
      assert(contexts_in_real[src->func_context - 1]->cache_valid);
      class_build(cache_lang_node_get(src->expr), contexts_in_real[src->func_context - 1], NULL, false, result,
          compiled);
      break;
    case vt_function:
    case vt_method:
      assert(!src->func_context || contexts_in_real[src->func_context - 1]);
      assert(!src->func_context || contexts_in_real[src->func_context - 1]->cache_valid);
      if(src->builtin)
        *result = builtin_functions[src->builtin - 1].value;
      else
        function_build(cache_lang_node_get(src->expr), contexts_in_real[src->func_context - 1], compiled, NULL, false,
            result);
      if(src->type == vt_function)
        break;
    case vt_string:
    case vt_object:
      result->string = cache_string_input.strings[src->string];
      if(src->type == vt_string)
        break;
      // FIXME Shouldn't the objects be somehow compacted?
      struct context *context = context_object_resolve(src->context - 1);
      result->object = context->object;
      break;
    default:
      assert(0);
  }
}

static void var_get(unsigned id, struct variable **vars_computed, size_t *var_values_pos)
{
  struct variable *var = var_alloc(NULL, 0, false);
  var->state = vs_cache;
  vars_computed[id] = var;
  var->cache_pos = *var_values_pos;
  var->count = vars_in->value_count;
  var->deps = vars_in->dep_list ? &dep_lists_in[vars_in->dep_list - 1] : NULL;
  *var_values_pos += vars_in->value_count;
}

static void dep_file_changed(struct var_dep *dep, struct var_dep_list *deps)
{
  if(!strcmp("", dep->filename))
  {
    if(config_cmdvars_changed)
      deps->changed = true;
  }
  else if(cached_file_changed(dep->filename) || config_hierarchy_rebuild)
    deps->changed = true;
}

static void cache_var_deps_resolve(struct var_dep_list *deps)
{
  deps->changed_resolved = true;
  deps->changed = false;
  dep_list_walk(deps, deps, (dep_list_walk_func) dep_file_changed);
}

void cache_context_resolve(void)
{
  dprinte("Resolving contexts");
  dep_lists_in = mempool_alloc(&cache_pool_input, dep_list_count_in * sizeof *dep_lists_in);
  for(size_t i = 0; i < dep_list_count_in; i ++)
  {
    for(size_t j = 0; j < dep_list_counts_in[i]; j ++)
    {
      struct var_dep *dep = mempool_alloc(&cache_pool_input, sizeof *dep);
      *dep = (struct var_dep) {
        .filename = cache_string_input.strings[*var_deps_in]
      };
      var_deps_in ++;
      dep_list_append(&dep_lists_in[i], dep);
    }
    dep_chain_append(dep_list_counts_in[i] == 1 ? &var_dep_single : &var_dep_merged, &dep_lists_in[i]);
    dep_lists_in[i].var_empty = var_empty;
    dep_lists_in[i].var_empty.deps = &dep_lists_in[i];
    dep_lists_in[i].var_undef = dep_lists_in[i].var_empty;
    dep_lists_in[i].var_undef.undefined = true;
  }
  unsigned last_id = 0;
  struct variable **vars_computed = mempool_alloc(&cache_pool_input, var_count * sizeof *vars_computed);
  struct var_holder **holders = mempool_alloc(&cache_pool_input, var_count * sizeof *holders);
  struct context *context;
  size_t var_values_pos = 0;
  contexts_in_real = mempool_alloc(&cache_pool_input, context_count * sizeof *contexts_in_real);
  for(size_t i = 0; i < context_count; i ++)
  {
    struct var_dep_list *deps = contexts_in->dep_list ? &dep_lists_in[contexts_in->dep_list - 1] : NULL;
    if(deps && !deps->changed_resolved)
      cache_var_deps_resolve(deps);
    bool valid = !deps || !deps->changed;
    if(valid)
    {
      // Create the context
      contexts_in_real[i] = context = context_get(&cache_pool_input);
      context_list_append(&input, context);
      context->dep_list = deps;
      context->filename = cache_string_input.strings[contexts_in->filename];
      context->token = contexts_in->token_number;
      context->cache_valid = true;
      context->node = cache_lang_node_get(contexts_in->class_code);
    }
    // Run trough variables
    while(contexts_in->var_count --)
    {
      if(vars_in->id > last_id) // New variable
        var_get(last_id ++, vars_computed, &var_values_pos);
      if(valid)
      {
        // Plug the variable into the context
        unsigned id = vars_in->id - 1;
        if(holders[id])
          context_var_reference(context, cache_string_input.strings[vars_in->name], holders[id], false);
        else
          holders[id] = context_var_set(context, cache_string_input.strings[vars_in->name], vars_computed[id], false,
              false); // TODO subscripted values?
      }
      vars_in ++;
    }
    contexts_in ++;
  }
  // Resolve parents between them
  contexts_in -= context_count;
  for(size_t i = 0; i < context_count; i ++)
    if(contexts_in_real[i])
    {
      if(contexts_in_real[i]->node)
        for(int j = contexts_in[i].parent_count - 1; j >= 0; j --)
        {
          assert(parents_in[j]);
          if(!contexts_in_real[parents_in[j] - 1])  // The parent is invalid, this one will get rebuilt anyway
            break;
          contexts_plug(contexts_in_real[parents_in[j] - 1], contexts_in_real[i]);
        }
      parents_in += contexts_in[i].parent_count;
    }
  for(size_t i = 0; i < context_count; i ++)
    if(contexts_in_real[i] && contexts_in_real[i]->node)
    {
      struct variable *target = variable_compute(contexts_in_real[i], "TARGET", &cache_pool_input_tmp);
      contexts_in_real[i]->cache_refs = target->count;
      overwrite_insert(contexts_in_real[i], false, &cache_pool_input_tmp);
    }
}

struct context_search
{
  struct context *result;
  const char *filename;
  unsigned token_num;
};

static void context_search(struct context *context, struct context_search *data)
{
  if(!data->result && context->filename && !strcmp(context->filename, data->filename) &&
      context->token == data->token_num)
    data->result = context;
}

struct context *cache_context_get(const char *filename, unsigned token_num)
{
  dprint("Searching for cached context %s/%u", filename, token_num);
  struct context_search data = {
    .filename = filename,
    .token_num = token_num
  };
  context_list_walk(&input, &data, (context_list_walk_func) context_search);
  if(data.result)
  {
    dprinte("Found it");
    context_list_remove(&input, data.result);
  }
  return data.result;
}

void cache_context_put(struct context *context, struct mempool_core *tmp_pool)
{
  assert(!context->cache_scheduled);
  context->cache_scheduled = true;
  dprint("Marking context %s/%u for cache storage", context->block->node->filename, context->block->node->token);
  context_list_append(&output, context);
  context->token = context->block->node->token;
  context->filename = context->block->node->filename;
  overwrite_insert(context, true, tmp_pool);
}

struct variable *cache_var_check(struct variable *variable, bool cmdline)
{
  assert(variable->state == vs_cache);
  if(variable->deps && !variable->deps->changed_resolved)
    cache_var_deps_resolve(variable->deps);
  if(cmdline || !variable->deps || !variable->deps->changed)
  {
    if(variable->count)
    {
      struct variable *new = var_alloc(NULL, variable->count, false);
      *new = *variable;
      for(size_t i = 0; i < new->count; i ++)
        var_value_load(&new->values[i], &var_values_in[new->cache_pos + i], new);
      new->state = vs_computed;
      return new;
    }
    else
    {
      variable->state = vs_computed;
      return variable;
    }
  }
  else
    return NULL;
}

void cache_context_init(void)
{
  overwrite_trie_init(&overwrite_trie, &persistent_pool);
}

struct context *cache_context_toplevel(void)
{
  return input.head;
}

void cache_context_restart(void)
{
  context_list_drop(&output);
  context_list_drop(&input);
  overwrite_trie = (struct overwrite_trie_t) {};
  context_count = var_count = var_value_count = parent_count_in = 0;
  contexts_in = NULL;
  vars_in = NULL;
  var_values_in = NULL;
  dep_list_count_in = 0;
  dep_list_counts_in = NULL;
  var_deps_in = NULL;
  dep_lists_in = NULL;
  contexts_in_real = NULL;
  parents_in = NULL;
}
