/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "files.h"
#include "global.h"
#include "strings.h"

#include <common/debug.h>
#include <common/file.h>
#include <common/config.h>

#include <assert.h>

/*
 * Fileinfo block description:
 * count (unsigned)
 * fileinfos, one after another
 */

static struct trie_t trie_input;
static bool trie_built;
static struct fileinfo *infos;
static unsigned count, count_out;

void cache_files_load(int fd)
{
  dprinte("Loading file stats from file");
  trie_init(&trie_input, &cache_pool_input);
  trie_built = false;
  cache_read(fd, &count, sizeof count);
  infos = mempool_alloc(&cache_pool_input, count * sizeof *infos);
  cache_read(fd, infos, count *sizeof *infos);
}

static void trie_build(void)
{
  trie_built = true;
  for(unsigned i = 0; i < count; i ++)
  {
    if(infos[i].filename >= cache_string_input.count)
    {
      diee("Cache corruption - string index out of range");
      // TODO exception
    }
    trie_insert(&trie_input, cache_string_input.strings[infos[i].filename], &infos[i]);
  }
}

static struct trie_t trie_mark;

static void mark(const char *filename)
{
  if(trie_find(&trie_mark, filename))
    return; // Already there
  trie_insert(&trie_mark, filename, &trie_mark);
  count_out ++;
}

void cache_file_mark(const char *filename)
{
  if(file_exists(filename))
    mark(filename);
}

bool cached_file_changed(const char *filename)
{
  assert(file_exists(filename));
  // TODO normalize the filename
  mark(filename);
  if(config_changed_all)
  {
    dprinte("All changed as said by user");
    return true;
  }
  if(!trie_built)
    trie_build();
  struct fileinfo *cached = trie_find(&trie_input, filename);
  if(!cached)
  {
    dprint("cached_file_changed: File not in cache (%s)", filename);
    return true;
  }
  struct fileinfo current = fileinfo_get(filename);
  return current.mod_time != cached->mod_time || current.size != cached->size;
}

static void write_walk(struct trie_node *node, const char *filename, const int *fd)
{
  unsigned idx = cache_string_push_dup(filename);
  struct fileinfo fi = fileinfo_get(filename);
  fi.filename = idx;
  cache_write(*fd, &fi, sizeof fi);
}

static void output_supply(struct trie_node *node, const char *filename, void *unused)
{
  if(file_exists(filename))
    mark(filename);
}

static void eliminate_missing(struct trie_node *node, const char *filename, void *unused)
{
  if(!file_exists(filename))
  {
    count_out --;
    node->data = NULL;
  }
}

void cache_files_store(int fd)
{
  dprinte("Storing file stats to cache");
  trie_walk(&trie_input, NULL, output_supply);
  trie_walk(&trie_mark, NULL, eliminate_missing);
  // Write the size
  cache_write(fd, &count_out, sizeof count_out);
  // Write the blocks
  trie_walk(&trie_mark, &fd, (trie_walk_func) write_walk);
  // Clean everything. Memory will be freed with the cache mempools
  count_out = 0;
}

void cache_files_init(void)
{
  trie_init(&trie_mark, &cache_pool_output);
}

void cache_files_restart(void)
{
  trie_input = (struct trie_t) {};
  count = count_out = 0;
  trie_built = false;
}
