/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "lang.h"
#include "global.h"

#include <lang/node.h>
#include <lang/block.h>
#include <lang/context.h>
#include <common/debug.h>
#include <common/config.h>

static size_t lang_tree_count, lang_node_count, block_scan_count,
              identifier_count, bind_count, ref_count;

static struct block_scan_result *scan_resolve(void *unused,
    struct block_scan_result *data);

#define DUMPER_PREFIX(X) dumper_node_##X
#define DUMPER_TYPE struct lang_node
#define DUMPER_INTERNAL_PTRS P(children_head) P(children_tail) P(next)
#define DUMPER_EXTERNAL_PTRS P(block_scan, scan_resolve)
#define DUMPER_STRINGS S(data) S(filename)
#define DUMPER_MEMPOOL
#include <common/dumper.h>

static struct dumper_node_state nodes_in;
#define NODE_FIND_IN(UNUSED, PTR) dumper_node_find(&nodes_in, PTR)

struct lang_tree_spec
{
  struct lang_tree_spec *next;
  struct lang_node *tree;
  const char *name;
};

#define DUMPER_PREFIX(X) dumper_tree_##X
#define DUMPER_TYPE struct lang_tree_spec
#define DUMPER_INTERNAL_PTRS P(next)
#define DUMPER_EXTERNAL_PTRS P(tree, NODE_FIND_IN)
#define DUMPER_STRINGS S(name)
#define DUMPER_MEMPOOL
#include <common/dumper.h>

static struct dumper_tree_state trees_in;

#define DUMPER_PREFIX(X) dumper_bind_##X
#define DUMPER_TYPE struct node_bind
#define DUMPER_INTERNAL_PTRS P(next)
#define DUMPER_EXTERNAL_PTRS P(node, NODE_FIND_IN)
#define DUMPER_MEMPOOL
#include <common/dumper.h>

static struct dumper_bind_state binds_in;
#define BIND_FIND_IN(UNUSED, PTR) dumper_bind_find(&binds_in, PTR)

#define DUMPER_PREFIX(X) dumper_identifier_##X
#define DUMPER_TYPE struct identifier
#define DUMPER_INTERNAL_PTRS P(next) P(symlink)
#define DUMPER_EXTERNAL_PTRS P(creators.head, BIND_FIND_IN)\
  P(creators.tail, BIND_FIND_IN)
#define DUMPER_STRINGS S(name)
#define DUMPER_MEMPOOL
#include <common/dumper.h>

static struct dumper_identifier_state ident_in;
#define ID_FIND_IN(UNUSED, PTR) dumper_identifier_find(&ident_in, PTR)

#define DUMPER_PREFIX(X) dumper_ref_##X
#define DUMPER_TYPE struct lang_ref
#define DUMPER_INTERNAL_PTRS P(list.head)
#define DUMPER_EXTERNAL_PTRS P(node, NODE_FIND_IN)
#define DUMPER_MEMPOOL
#include <common/dumper.h>

static struct dumper_ref_state refs_in;
#define REF_FIND_IN(UNUSED, PTR) dumper_ref_find(&refs_in, PTR)

#define DUMPER_PREFIX(X) dumper_scan_##X
#define DUMPER_TYPE struct block_scan_result
#define ID_LIST(var) P(var.head, ID_FIND_IN) P(var.tail, ID_FIND_IN)
#define DUMPER_EXTERNAL_PTRS P(parent_file, NODE_FIND_IN) P(node, NODE_FIND_IN) \
  ID_LIST(mentioned) ID_LIST(out) P(sub_files.head, REF_FIND_IN) P(sub_blocks.head, REF_FIND_IN)
#define DUMPER_MEMPOOL
#include <common/dumper.h>

static struct dumper_scan_state scans_in;

static struct lang_tree_spec_list
{
  struct lang_tree_spec *head, *tail;
} tree_spec_list;

#define LIST_PREFIX(X) tree_list_##X
#define LIST_NODE struct lang_tree_spec
#define LIST_CONTAINER struct lang_tree_spec_list
#define LIST_NEXT next
#define LIST_HEAD head
#define LIST_TAIL tail
#include <common/list.h>

#define MEMPOOL_PREFIX(X) lang_tree_pool_##X
#define MEMPOOL_TYPE struct lang_tree_spec
#define MEMPOOL_HARDCODE_VARIABLE &cache_pool_output
#include <common/mempool.h>

void cache_lang_init(void)
{
  tree_list_drop(&tree_spec_list);
}

struct build_dumpers
{
  struct dumper_tree_state trees;
  struct dumper_scan_state scans;
  struct dumper_node_state nodes;
  struct dumper_identifier_state identifiers;
  struct dumper_bind_state binds;
  struct dumper_ref_state refs;
};

static void build_dumpers_binds(struct node_bind *bind, struct build_dumpers *dumpers)
{
  dumper_bind_push(&dumpers->binds, bind);
}

static void build_dumpers_identifiers(struct identifier *id, struct build_dumpers *dumpers)
{
  dumper_identifier_push(&dumpers->identifiers, id);
  node_bind_walk(&id->creators, dumpers, (node_bind_walk_func) build_dumpers_binds);
}

static void build_dumpers_refs(struct lang_ref *ref, struct build_dumpers *dumpers)
{
  dumper_ref_push(&dumpers->refs, ref);
}

static void build_dumpers_nodes(struct lang_node *node, struct build_dumpers *dumpers)
{
  dumper_node_push(&dumpers->nodes, node);
  if(node->block_scan)
  {
    dumper_scan_push(&dumpers->scans, node->block_scan);
    identifier_list_walk(&node->block_scan->mentioned, dumpers, (identifier_list_walk_func) build_dumpers_identifiers);
    identifier_list_walk(&node->block_scan->out, dumpers, (identifier_list_walk_func) build_dumpers_identifiers);
    lang_ref_list_walk(&node->block_scan->sub_blocks, dumpers, (lang_ref_list_walk_func) build_dumpers_refs);
    lang_ref_list_walk(&node->block_scan->sub_files, dumpers, (lang_ref_list_walk_func) build_dumpers_refs);
  }
  langlist_walk(node, dumpers, (langlist_walk_func) build_dumpers_nodes);
}

static void build_dumpers(struct lang_tree_spec *tree, struct build_dumpers *dumpers)
{
  dumper_tree_push(&dumpers->trees, tree);
  build_dumpers_nodes(tree->tree, dumpers);
}

void cache_lang_store(int fd)
{
  dprinte("Storing lang trees");
  struct build_dumpers dumpers;
  struct mempool_core tmp_pool;
  mempool_core_init(&tmp_pool, 0, true);
  dumper_tree_init(&dumpers.trees, lang_tree_count, &tmp_pool);
  dumper_scan_init(&dumpers.scans, block_scan_count, &tmp_pool);
  dumper_node_init(&dumpers.nodes, lang_node_count, &tmp_pool);
  dumper_identifier_init(&dumpers.identifiers, identifier_count, &tmp_pool);
  dumper_bind_init(&dumpers.binds, bind_count, &tmp_pool);
  dumper_ref_init(&dumpers.refs, ref_count, &tmp_pool);
  tree_list_walk(&tree_spec_list, &dumpers, (tree_list_walk_func) build_dumpers);
  dumper_bind_check(&dumpers.binds);
  assert(dumpers.trees.used == dumpers.trees.size);
  cache_write(fd, &lang_tree_count, sizeof lang_tree_count);
  cache_write(fd, &dumpers.scans.size, sizeof dumpers.scans.size);
  cache_write(fd, &dumpers.nodes.size, sizeof dumpers.nodes.size);
  cache_write(fd, &dumpers.identifiers.size, sizeof dumpers.identifiers.size);
  cache_write(fd, &dumpers.binds.used, sizeof dumpers.binds.used);
  cache_write(fd, &dumpers.refs.used, sizeof dumpers.refs.used);
  cache_write(fd, dumpers.trees.data, lang_tree_count * sizeof *dumpers.trees.data);
  cache_write(fd, dumpers.scans.data, dumpers.scans.size * sizeof *dumpers.scans.data);
  cache_write(fd, dumpers.nodes.data, dumpers.nodes.size * sizeof *dumpers.nodes.data);
  cache_write(fd, dumpers.identifiers.data, dumpers.identifiers.size * sizeof *dumpers.identifiers.data);
  cache_write(fd, dumpers.binds.data, dumpers.binds.used * sizeof *dumpers.binds.data);
  cache_write(fd, dumpers.refs.data, dumpers.refs.used * sizeof *dumpers.refs.data);
  mempool_core_clean(&tmp_pool);
}

static struct block_scan_result *scan_resolve(void *unused,
    struct block_scan_result *ptr)
{
  return dumper_scan_find(&scans_in, ptr);
}

void cache_lang_load(int fd)
{
  dprinte("Loading lang trees");
  size_t lang_tree_count_in, block_scan_count_in, lang_node_count_in, identifier_count_in, bind_count_in, ref_count_in;
  cache_read(fd, &lang_tree_count_in, sizeof lang_tree_count_in);
  cache_read(fd, &block_scan_count_in, sizeof block_scan_count_in);
  cache_read(fd, &lang_node_count_in, sizeof lang_node_count_in);
  cache_read(fd, &identifier_count_in, sizeof identifier_count_in);
  cache_read(fd, &bind_count_in, sizeof bind_count_in);
  cache_read(fd, &ref_count_in, sizeof ref_count_in);
  dumper_tree_init(&trees_in, lang_tree_count_in, &cache_pool_input);
  dumper_scan_init(&scans_in, block_scan_count_in, &cache_pool_input);
  dumper_node_init(&nodes_in, lang_node_count_in, &cache_pool_input);
  dumper_identifier_init(&ident_in, identifier_count_in, &cache_pool_input);
  dumper_bind_init(&binds_in, bind_count_in, &cache_pool_input);
  dumper_ref_init(&refs_in, ref_count_in, &cache_pool_input);
  cache_read(fd, trees_in.data, lang_tree_count_in * sizeof *trees_in.data);
  cache_read(fd, scans_in.data, block_scan_count_in * sizeof *scans_in.data);
  cache_read(fd, nodes_in.data, lang_node_count_in * sizeof *nodes_in.data);
  cache_read(fd, ident_in.data, identifier_count_in * sizeof *ident_in.data);
  cache_read(fd, binds_in.data, bind_count_in * sizeof *binds_in.data);
  cache_read(fd, refs_in.data, ref_count_in * sizeof *refs_in.data);
  trees_in.used = trees_in.size;
  scans_in.used = scans_in.size;
  nodes_in.used = nodes_in.size;
  ident_in.used = ident_in.size;
  binds_in.used = binds_in.size;
  refs_in.used = refs_in.size;
}

void cache_lang_resolve(void)
{
  dprinte("Resolving pointers in loaded lang trees");
  dumper_tree_resolve(&trees_in);
  dumper_scan_resolve(&scans_in);
  dumper_node_resolve(&nodes_in);
  dumper_identifier_resolve(&ident_in);
  dumper_bind_resolve(&binds_in);
  dumper_ref_resolve(&refs_in);
}

void cache_lang_out_free(void)
{
  lang_tree_count = lang_node_count = block_scan_count =
    identifier_count = bind_count = ref_count = 0;
  tree_list_drop(&tree_spec_list);
}

static void context_resolve(struct lang_ref *ref, bool *toplevel)
{
  struct lang_node *node = ref->node;
  if(node->type != ln_block || !node->block_scan->global_context)
    return;
  context_raw_block(node->block_scan, node);
  if(toplevel && *toplevel)
    contexts_plug(context_toplevel, node->block_scan->global_context);
  lang_ref_list_walk(&node->block_scan->sub_blocks, NULL, (lang_ref_list_walk_func) context_resolve);
}

struct lang_node *cache_lang_tree_get(const char *name, bool toplevel)
{
  if(config_hierarchy_rebuild)
    return NULL;
  for(size_t i = 0; i < trees_in.used; i ++)
  {
    if(!strcmp(name, trees_in.data[i].data.name))
    {
      context_resolve(&(struct lang_ref) { .node = trees_in.data[i].data.tree }, &toplevel);
      dprint("Extract langtree %p (%s)", trees_in.data[i].data.tree, name);
      cache_lang_tree_push(trees_in.data[i].data.tree, name);
      return trees_in.data[i].data.tree;
    }
  }
  return NULL;
}

static void binds_count(struct node_bind *b, void *unused)
{
  bind_count ++;
}

static void idents_count(struct identifier *id, void *unused)
{
  identifier_count ++;
  node_bind_walk(&id->creators, NULL, binds_count);
}

static void refs_count(struct lang_ref *ref, void *unused)
{
  ref_count ++;
}

static void nodes_count(struct lang_node *node, void *unused)
{
  lang_node_count ++;
  langlist_walk(node, NULL, nodes_count);
  if(node->block_scan && node->type == ln_block)
  {
    block_scan_count ++;
    identifier_list_walk(&node->block_scan->mentioned, NULL, idents_count);
    identifier_list_walk(&node->block_scan->out, NULL, idents_count);
    lang_ref_list_walk(&node->block_scan->sub_blocks, NULL, refs_count);
    lang_ref_list_walk(&node->block_scan->sub_files, NULL, refs_count);
  }
}

void cache_lang_tree_push(struct lang_node *tree, const char *name)
{
  dprint("Marking tree for cache dump, walking it to count nodes (%p, %p, %s)", tree, tree->block_scan, name);
  lang_tree_count ++;
  struct lang_tree_spec *spec = lang_tree_pool_get();
  spec->name = strpool_dup(&strpool, name);
  spec->tree = tree;
  tree_list_append(&tree_spec_list, spec);
  nodes_count(tree, NULL);
}

struct lang_node *cache_lang_node_get(struct lang_node *old_ptr)
{
  return dumper_node_find(&nodes_in, old_ptr);
}

void cache_lang_restart(void)
{
  lang_tree_count = lang_node_count = block_scan_count = identifier_count = bind_count = ref_count = 0;
  nodes_in = (struct dumper_node_state) {};
  trees_in = (struct dumper_tree_state) {};
  binds_in = (struct dumper_bind_state) {};
  ident_in = (struct dumper_identifier_state) {};
  refs_in = (struct dumper_ref_state) {};
  scans_in = (struct dumper_scan_state) {};
  tree_list_drop(&tree_spec_list);
}
