/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "strings.h"
#include "global.h"

#include <common/memory.h>
#include <common/debug.h>

#include <stdbool.h>

/*
 * String block description:
 * Count of strings (unsigned)
 * Number of bytes the strings take (size_t)
 * Strings, one after another, each terminating by '\0'
 */

// Loading

struct cache_string_block cache_string_input;

void cache_string_clean(void)
{
  cache_string_input.count = 0;
}

void cache_string_load(int fd)
{
  dprinte("Loading strings from cache");
  cache_string_clean();
  cache_read(fd, &cache_string_input.count, sizeof cache_string_input.count);
  cache_string_input.count ++;
  size_t size;
  cache_read(fd, &size, sizeof size);
  if(size)
  {
    char *strings = mempool_alloc(&cache_pool_input, size);
    cache_read(fd, strings, size);
    // All the strings are in the same memory block, one after another
    cache_string_input.strings = mempool_alloc(&cache_pool_input,
        cache_string_input.count * sizeof *cache_string_input.strings);
    // The following part gets pointers to start of each string and puts them into the array
    const char *end_str = strings + size;
    const char **end_arr = cache_string_input.strings + cache_string_input.count;
    const char **curr = cache_string_input.strings + 1;
    *curr ++ = strings ++;// Get the first string
    while(strings < end_str && curr < end_arr)
    {
      if(!*strings ++) // One ended, mark the next one
        *curr ++ = strings;
    }
    // Just check consistency
    while(strings < end_str && *strings ++);
    if(curr != end_arr || strings != end_str || *(strings - 1))
      diee("String cache corruption"); // TODO replace die with exception
  }
  else
    cache_string_input.strings = NULL;
}

// Storing

#define STACK_TYPE const char *
#define STACK_PREFIX(X) output_##X
#include <common/stack.h>

static char base;
static struct trie_t output_trie;

static struct output_stack output;

void cache_string_fill_start(void)
{
  assert(!output.size);
}

static unsigned cache_string_push_internal(const char *str)
{
  output_push(&output, str);
  trie_insert(&output_trie, str, &base + output.size - 1);
  return output.size - 1;
}

static bool lookup(const char *str, unsigned *result)
{
  if(!str)
  {
    *result = 0;
    return true;
  }
  char *idx = trie_find(&output_trie, str);
  if(idx)
  {
    *result = idx - &base;
    return true;
  }
  else
  {
    return false;
  }
}

unsigned cache_string_push(const char *str)
{
  unsigned result;
  if(lookup(str, &result))
    return result;
  return cache_string_push_internal(str);
}

unsigned cache_string_push_dup(const char *str)
{
  unsigned result;
  if(lookup(str, &result))
    return result;
  return cache_string_push_internal(strpool_dup(&cache_strpool_output, str));
}

void cache_string_store_drop(int fd)
{
  dprinte("Storing strings to cache");
  // Store count
  unsigned count = output.size - 1;
  cache_write(fd, &count, sizeof count);
  // Compute and store total size of strings
  size_t size = 0;
  size_t sizes[count + 1];
  for(size_t i = 1; i < output.size; i ++)
    size += sizes[i] = 1 + strlen(output.data[i]);
  cache_write(fd, &size, sizeof size);
  // Strote the strings
  for(size_t i = 1; i < output.size; i ++)
    cache_write(fd, output.data[i], sizes[i]);
  // Free everything (The stringpool and trie pool will be freed globaly in the cache)
  output_clean(&output);
}

// Init

void cache_string_init(void)
{
  cache_string_input.count = 0;
  cache_string_input.strings = NULL;
  output_init(&output);
  output_push(&output, NULL);
  trie_init(&output_trie, &cache_pool_output);
}

void cache_string_restart(void)
{
  output_trie = (struct trie_t) {};
  output_free(&output);
  output = (struct output_stack) {};
  cache_string_input = (struct cache_string_block) {};
}
