/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "debug.h"
#include "config.h"
#include "stringpool.h"

#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

enum args
{
  ar_cache = 1024,
  ar_nocache,
  ar_def_ext,
  ar_filename,
  ar_hier_rebuild
};

struct option_ex {
  struct option opt;
  const char *help;
};

static struct option_ex args[] = {
  { { "file", 1, NULL, 'f' }, "Filename to read instructions from." },
  { { "default-file", 1, NULL, ar_filename },
    "What is default file name for instruction file when entering directory (default REM)" },
  { { "cache", 1, NULL, ar_cache }, "Use file as cache (default .rem_cache)" },
  { { "nocache", 0, NULL, ar_nocache }, "Do not read cache" },
  { { "import", 1, NULL, 'I' }, "Add directory to search for imports. Can be provided multiple times" },
  { { "define", 1, NULL, 'D' }, "Provide a variable content." },
  { { "targets", 1, NULL, 't' }, "What variable should be used as target list (default TARGETS)" },
  { { "method", 1, NULL, 'm' }, "What method to execute on targets. Can be provided multiple times" },
  { { "all-run", 0, NULL, 'r' }, "Run all compilation actions, no matter if deps changed" },
  { { "all-changed", 0, NULL, 'c' }, "Consider all files modified since last time" },
  //{ { "define-ext", 1, NULL, ar_def_ext }, "Extended syntax TODO
  { { "hierarchy-rebuild", 0, NULL, ar_hier_rebuild }, "Rebuild hierarchy of modules, do not load them from cache" },
  { { "help", 0, NULL, 'h' }, "This help text" },
#ifdef DEBUG
  { { "debug", 0, NULL, 'd' }, "Print lots of annoying messages of internal character" },
#endif
  { { NULL, 0, NULL, 0 }, NULL }
};

static const char *short_args = "f:m:t:I:rcD:h"
#ifdef DEBUG
"d"
#endif
;

const char *conf_filename = "REM";
const char *conf_filename_start = "REM";
const char *cache_filename = ".rem_cache";
const char *config_targets = "TARGETS";
struct strlist config_methods;
struct strlist config_import_path_list;
bool nocache_flag;
bool config_run_all;
bool config_changed_all;
bool config_cmdvars_changed;
bool config_hierarchy_rebuild;
bool config_debug;
struct strlist config_defines;
struct mempool_core config_pool;

static void import_insert(char *string, void (*function)(struct strlist *, struct strlist *))
{
  struct strlist *new = mempool_alloc(&config_pool, sizeof *new);
  new->string = string;
  function(&config_import_path_list, new);
}

#ifdef TRACE
static void strlist_print(struct strlist *list, void *unused)
{
  fputs(list->string, stderr);
  fputc('\n', stderr);
}
#endif

void config_init(int argc, char *argv[], struct mempool_core *tmp_pool)
{
  mempool_core_init(&config_pool, 0, false);
  char *env;
  if((env = getenv("REMAKE_FILENAME")))
    conf_filename = conf_filename_start = env;
  if((env = getenv("REMAKE_CACHE")))
    cache_filename = env;
  if((env = getenv("REMAKE_TARGETS")))
    config_targets = env;
  if((env = getenv("REMAKE_IMPORT_PATHS")))
  {
    env = strpool_dup(&strpool, env);
    char *path;
    while((path = strtok(env, ":")))
    {
      env = NULL;
      import_insert(path, strlist_append);
    }
  }
  if((env = getenv("HOME")))
  {
    size_t len = strlen(env);
    const char suffix[] = "/.remake/modules";
    char *path = mempool_alloc(&config_pool, len + sizeof suffix);
    memcpy(path, env, len);
    memcpy(path + len, suffix, sizeof suffix);
    import_insert(path, strlist_append);
  }
  import_insert("/etc/remake/modules", strlist_append);
  import_insert("remake_modules", strlist_prepend);
  size_t cnt = 0;
  for(struct option_ex *oe = args; oe->help; oe ++)
    cnt ++;
  struct option args_core[cnt + 1];
  for(size_t i = 0; i <= cnt; i ++)
    args_core[i] = args[i].opt;
  int option;
  while((option = getopt_long(argc, argv, short_args, args_core, NULL)) >= 0)
  {
    switch(option)
    {
      case ar_filename:
        conf_filename = optarg;
        break;
      case 'f':
        conf_filename_start = optarg;
        break;
      case ar_cache:
        cache_filename = optarg;
        break;
      case 't':
        config_targets = optarg;
        break;
      case 'm':
      {
        struct strlist *new = mempool_alloc(&config_pool, sizeof *new);
        new->string = optarg;
        strlist_append(&config_methods, new);
        break;
      }
      case ar_nocache:
        nocache_flag = true;
        break;
      case 'I':
        import_insert(optarg, strlist_prepend);
        break;
      case 'r':
        config_run_all = true;
        break;
      case 'c':
        config_changed_all = true;
        break;
      case ar_hier_rebuild:
        config_hierarchy_rebuild = true;
        break;
      case 'd':
        config_debug = true;
        break;
      case 'D':
      case ar_def_ext:
      {
        struct strlist *new = mempool_alloc(&config_pool, sizeof *new);
        new->string = optarg;
        new->common = (option == 'D');
        strlist_append(&config_defines, new);
        config_cmdvars_changed = true;
        break;
      }
      case 'h':
      case '?':
        fputs("remake [options]:\n", stderr);
        for(struct option_ex *oe = args; oe->help; oe ++)
          fprintf(stderr, " %c%c%c--%-18s %s\n", oe->opt.val < 256 ? '-' : ' ', oe->opt.val < 256 ? oe->opt.val : ' ',
              oe->opt.val < 256 ? ',' : ' ', oe->opt.name, oe->help);
        exit(1);
        break;
    }
  }
#ifdef TRACE
  if(config_debug)
  {
    dprinte("Paths to search for modules:");
    strlist_walk(&config_import_path_list, NULL, strlist_print);
  }
#endif
}
