/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "mempool_core.h"
#include "memory.h"
#include "pagepool.h"
#include "debug.h"
#include "restart.h"

#include <string.h>

struct pool_page
{
  struct pool_page *next;
  size_t size;
  unsigned char data[];
};

#define PAGE_REST (PAGE_SIZE - sizeof(struct pool_page))

#define LIST_PREFIX(X) list_##X
#define LIST_CONTAINER struct mempool_core
#define LIST_NODE struct pool_page
#define LIST_NEXT next
#define LIST_HEAD first
#include "list.h"

void mempool_core_init(struct mempool_core *mcore, size_t align, bool auto_cleanup)
{
  mcore->align = align ? align : ALIGN_SIZE;
  mcore->empty = 0;
  mcore->borrowed = 0;
  mcore->references = 0;
  mcore->auto_cleanup = auto_cleanup;
  mcore->cleanup = NULL;
  list_drop(mcore);
}

static void page_drop(struct pool_page *page, void *unused)
{
  page_return(page, page->size);
}

void mempool_core_clean(struct mempool_core *mcore)
{
  if(mcore->cleanup)
  {
    restart_cleanup_unregister(mcore->cleanup);
    mcore->cleanup = NULL;
  }
  list_walk(mcore, NULL, page_drop);
  list_drop(mcore);
  mcore->empty = 0;
}

// Round to nearest bigger page size
static size_t page_round(size_t size)
{
  return PAGE_SIZE + ((size - PAGE_REST) + PAGE_SIZE - 1 ) /
    PAGE_SIZE * PAGE_SIZE;
}

static void cleanup_init(struct mempool_core *pool)
{
  if(pool->auto_cleanup && !pool->cleanup)
    pool->cleanup = restart_cleanup_register((restart_cleanup_func) mempool_core_clean, pool);
}

static void *page_plug(struct mempool_core *pool, size_t size, size_t psize)
{
  dprint("Getting new page(s), %zu, pool: %p", psize, pool);
  cleanup_init(pool);
  struct pool_page *page = page_alloc(psize);
#ifndef CAN_MMAP
  memset(page, 0, psize);
#endif
  page->size = psize;
  list_prepend(pool, page);
  if(psize - size < pool->empty)
  {
    return page->data;
  }
  else
  {
    pool->empty = psize - sizeof *page;
    pool->data = page->data;
  }
  return NULL;
}

void *mempool_alloc_nozero(struct mempool_core *pool, size_t size)
{
#ifdef MEMPOOL_DUMMY
  void *data = xmalloc(size);
  memset(data, 0, size);
  return data;
#endif
  void *result;
  if(size > pool->empty)
  {
    if(result = page_plug(pool, size, page_round(size)))
      return result;
  }
  result = pool->data;
  // Keep alignment
  size_t diff = (size + pool->align - 1) / pool->align * pool->align;
  pool->data += diff;
  pool->empty -= diff;
  pool->borrowed = pool->borrowed > diff ? pool->borrowed - diff : 0;
  return result;
}

void *mempool_alloc(struct mempool_core *pool, size_t size)
{
  if(pool->borrowed)
  {
    memset(pool->data, 0, pool->borrowed);
    pool->borrowed = 0;
  }
  return mempool_alloc_nozero(pool, size);
}

struct mempool_core persistent_pool;

void mempool_src_init(void)
{
  mempool_core_init(&persistent_pool, 0, true);
}

void *mempool_alloc_resizeable(struct mempool_core *mcore, size_t *size)
{
  cleanup_init(mcore);
  struct pool_page *page = page_alloc(PAGE_SIZE);
#ifndef CAN_MMAP
  memset(page, 0, PAGE_SIZE);
#endif
  *size = PAGE_REST;
  page->size = PAGE_SIZE;
  list_prepend(mcore, page);
  return page->data;
}

void *mempool_resize_page(struct mempool_core *mempool,
    struct pool_page *page, size_t size)
{
  struct pool_page *new_page = xrealloc_mmap(page, page->size, size);
  new_page->size = size;
  if(new_page != page)
  {
    list_remove(mempool, page);
    list_append(mempool, new_page);
  }
  return new_page;
}

void *mempool_grow(struct mempool_core *mempool, void *data,
    size_t *new_size)
{
  char *dpos = data;
  dpos -= sizeof(struct pool_page);
  struct pool_page *page = (struct pool_page *) dpos;
  *new_size = page->size * 2;
  struct pool_page *new_page = mempool_resize_page(mempool, page,
      *new_size);
  *new_size -= sizeof(struct pool_page);
  return new_page->data;
}

void *mempool_size_set(struct mempool_core *mempool, void *data,
    size_t size)
{
  char *dpos = data;
  dpos -= sizeof(struct pool_page);
  struct pool_page *page = (struct pool_page *) dpos;
  if(size <= page->size - sizeof(struct pool_page))
  {
    if(size <= mempool->empty)
    {
      memcpy(mempool->data, page->data, size);
      list_remove(mempool, page);
      xfree_mmap(page, page->size);
      return mempool_alloc(mempool, size);
    }
    if(page->size - size - sizeof(struct pool_page) > mempool->empty)
    {
      mempool->data = page->data + size;
      mempool->empty = page->size - size - sizeof(struct pool_page);
#ifndef CAN_MMAP
      memset(mempool->data, 0, mempool->empty);
#endif
      return page->data;
    }
  }
  struct pool_page *new_page = mempool_resize_page(mempool, page,
      size + sizeof(struct pool_page));
  return new_page->data;
}

void *mempool_borrow(struct mempool_core *pool, size_t size)
{
  if(pool->empty < size)
  {
    void *result = page_plug(pool, 0, page_round(size));
    assert(!result);  // The new one should be bigger than the old one
  }
  pool->borrowed = (size > pool->borrowed) ? size : pool->borrowed;
  return pool->data;
}
