/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "restart.h"
#include "mempool_core.h"
#include "config.h"
#include "dir.h"
#include "stringpool.h"
#include "debug.h"

#include <strings.h>

struct restart_cleanup
{
  restart_cleanup_func func;
  void *data;
  struct restart_cleanup *next, *prev;
  bool active:1;
};

struct restart_cleanup_container
{
  struct restart_cleanup *head, *tail;
};

static bool inited, active = true;
struct mempool_core restart_pool;

#define LIST_PREFIX(X) restart_cleanup_##X
#define LIST_NODE struct restart_cleanup
#define LIST_CONTAINER struct restart_cleanup_container
#define LIST_NEXT next
#define LIST_PREV prev
#define LIST_HEAD head
#define LIST_TAIL tail
#include "list.h"

#define RECYCLER_PREFIX(X) restart_cleanup_recycler_##X
#define RECYCLER_TYPE struct restart_cleanup
#define RECYCLER_NEXT next
#include "recycler.h"

static const char *dir;
static struct restart_cleanup_container cleanups;
static struct restart_cleanup_recycler_t recycler;

jmp_buf restart_context;

static void make_init(void)
{
  if(!inited)
  {
    inited = true;
    mempool_core_init(&restart_pool, 0, false);
    restart_cleanup_recycler_init(&recycler, &restart_pool);
  }
}

static void cleanup_perform(struct restart_cleanup *cleanup, void *unused)
{
  cleanup->active = false;
  cleanup->func(cleanup->data);
  restart_cleanup_recycler_put(&recycler, cleanup);
}

void restart_use(const char *file)
{
  dprint("Jump to start with file %s", file ? file : "");
  make_init();
  if(file) // Do we go elsewhere
  {
    char *f = strpool_dup(&restart_pool, file);
    char *slash = rindex(f, '/');
    if(slash)
    {
      conf_filename_start = strpool_dup(&config_pool, ++ slash);
      *slash = '\0';
      dir = f;
    }
    else
    {
      conf_filename_start = strpool_dup(&config_pool, f);
      dir = NULL;
    }
    if(!conf_filename_start || !*conf_filename_start)
      conf_filename_start = conf_filename;
  }
  make_init();
  restart_cleanup_walk(&cleanups, NULL, cleanup_perform);
  restart_cleanup_drop(&cleanups);
  restart_cleanup_recycler_clean(&recycler);
  longjmp(restart_context, 1);
}

void restart_done(void)
{
  if(dir)
    dir_set(dir);
  dir = NULL;
  mempool_core_clean(&restart_pool);
  dprinte("Jump done");
}

void restart_finish(void)
{
  if(inited)
    mempool_core_clean(&restart_pool);
  active = false;
}

struct restart_cleanup *restart_cleanup_register(restart_cleanup_func func, void *data)
{
  if(!active)
    return NULL;
  make_init();
  struct restart_cleanup *cleanup = restart_cleanup_recycler_get(&recycler);
  *cleanup = (struct restart_cleanup) {
    .func = func,
    .data = data,
    .active = true
  };
  restart_cleanup_append(&cleanups, cleanup);
}

void restart_cleanup_unregister(struct restart_cleanup *cleanup)
{
  if(!active)
    return;
  if(cleanup->active)
  {
    restart_cleanup_remove(&cleanups, cleanup);
    cleanup->active = false;
    restart_cleanup_recycler_put(&recycler, cleanup);
  }
}
