/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "str.h"
#include "mempool_core.h"
#include <stdbool.h>
#include <string.h>
#define TRIE_NODE TRIE_PREFIX(node)
#define TRIE_T TRIE_PREFIX(t)

struct TRIE_NODE
{
  TRIE_DATA data;
  const char *text;
  size_t length;
  struct TRIE_NODE *next, *head;
};

struct TRIE_T
{
  size_t depth;
  struct mempool_core *pool;
  struct TRIE_NODE root;
};

static void TRIE_PREFIX(init)(struct TRIE_T *trie, struct mempool_core *pool)
{
  *trie = (struct TRIE_T) {
#ifdef TRIE_DATA_DEFAULT
    .root.data = TRIE_DATA_DEFAULT,
#endif
    .pool = pool
  };
}

static struct TRIE_NODE *TRIE_PREFIX(new)(struct TRIE_T *trie, const char *name)
{
  struct TRIE_NODE *result = mempool_alloc(trie->pool, sizeof *result);
  char *data;
  result->length = name ? strlen(name) : 0;
  result->text = data = mempool_alloc(trie->pool, result->length);
  memcpy(data, name, result->length);
#ifdef TRIE_DATA_DEFAULT
  result->data = TRIE_DATA_DEFAULT;
#endif
  return result;
}

// The magic for walking & creating the trie. The rest are just wrappers for this monster
static struct TRIE_NODE *TRIE_PREFIX(path)(struct TRIE_T *trie, const char *string, bool create)
{
  struct TRIE_NODE *node = &trie->root;
  while(*string) // Still somewhere to search
  {
    for(struct TRIE_NODE **sub = &node->head; *sub; sub = &(*sub)->next) // Walk branches
      if((*sub)->length && *(*sub)->text == *string) // Starting alright
      {
        const char *d = strfdcn((*sub)->text, string, (*sub)->length);
        if(d) // They start the same, but differ somewhere.
        {
          if(create)
          { // Duh. Split.
            // Create the middle one
            struct TRIE_NODE *middle = TRIE_PREFIX(new)(trie, NULL);
            // First half of the text here
            middle->text = (*sub)->text;
            middle->length = d - (*sub)->text;
            // The rest there
            (*sub)->length -= middle->length;
            (*sub)->text = d;
            // Put it into the place of the old one
            middle->head = *sub;
            middle->next = (*sub)->next;
            (*sub)->next = NULL;
            *sub = middle;
            // Move to the middle one
            node = middle;
            string += middle->length;
            if(!*string)
              return middle;
            // No need to search more, jump to create the new node
            break;
          }
          else
            return NULL;
        }
        else
        {
          string += (*sub)->length;
          node = *sub;
          goto NEXT_RUN;
        }
      }
    // Not here at all.
    if(create) // New one, starting here
    {
      struct TRIE_NODE *new = TRIE_PREFIX(new)(trie, string);
      new->next = node->head;
      node->head = new;
      return new;
    }
    else
      return NULL;
    NEXT_RUN:;
  }
  return node;
}

static TRIE_DATA TRIE_PREFIX(find_default)(struct TRIE_T *trie, const char *string, TRIE_DATA def)
{
  struct TRIE_NODE *node = TRIE_PREFIX(path)(trie, string, false);
  if(node)
    return node->data;
  else
    return def;
}

#ifdef TRIE_DATA_DEFAULT
static TRIE_DATA TRIE_PREFIX(find)(struct TRIE_T *trie, const char *string)
{
  return TRIE_PREFIX(find_default)(trie, string, TRIE_DATA_DEFAULT);
}
#endif

static void TRIE_PREFIX(insert)(struct TRIE_T *trie, const char *string, TRIE_DATA value)
{
  size_t len = strlen(string);
  if(len > trie->depth)
    trie->depth = len;
  TRIE_PREFIX(path)(trie, string, true)->data = value;
}

typedef void (*TRIE_PREFIX(walk_func))(struct TRIE_NODE *node, const char *string, void *data);

static void TRIE_PREFIX(walk_internal)(struct TRIE_NODE *node, void *data, TRIE_PREFIX(walk_func) func, char *path,
    char *ppos, bool always)
{
  for(struct TRIE_NODE *sub = node->head; sub; sub = sub->next)
  {
    memcpy(ppos, sub->text, sub->length);
    TRIE_PREFIX(walk_internal)(sub, data, func, path, ppos + sub->length, always);
  }
  if(always || node->data != TRIE_DATA_DEFAULT)
  {
    *ppos = '\0';
    func(node, path, data);
  }
}

static void TRIE_PREFIX(walk_switch)(struct TRIE_T *trie, void *data, TRIE_PREFIX(walk_func) func, bool always)
{
  char path[trie->depth + 1];
  TRIE_PREFIX(walk_internal)(&trie->root, data, func, path, path, always);
}

static void TRIE_PREFIX(walk)(struct TRIE_T *trie, void *data, TRIE_PREFIX(walk_func) func)
{
  TRIE_PREFIX(walk_switch)(trie, data, func, false);
}

static void TRIE_PREFIX(walk_all)(struct TRIE_T *trie, void *data, TRIE_PREFIX(walk_func) func)
{
  TRIE_PREFIX(walk_switch)(trie, data, func, true);
}

static void TRIE_PREFIX(size_func)(struct TRIE_NODE *node, const char *string, size_t *count)
{
  (*count) ++;
}

static size_t TRIE_PREFIX(size)(struct TRIE_T *trie)
{
  size_t result;
  TRIE_PREFIX(walk)(trie, &result, (TRIE_PREFIX(walk_func)) TRIE_PREFIX(size_func));
  return result;
}

#undef TRIE_DATA
#undef TRIE_DATA_DEFAULT
#undef TRIE_NODE
#undef TRIE_NODE_ALLOC
#undef TRIE_PREFIX
#undef TRIE_T
