/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "hierarchy.h"

#include <common/debug.h>
#include <common/config.h>
#include <common/stringpool.h>
#include <common/file.h>
#include <common/restart.h>
#include <lang/node.h>
#include <lang/block.h>
#include <lang/compute.h>
#include <lang/var.h>
#include <lang/builtin.h>
#include <lang/context.h>
#include <parser/parser.h>
#include <cache/cache.h>
#include <cache/files.h>
#include <cache/lang.h>

#include <string.h>
#include <strings.h>
#include <signal.h>
#include <errno.h>
#include <setjmp.h>
#include <stdio.h>

#define RECYCLER_PREFIX(X) ref_rec_##X
#define RECYCLER_TYPE struct lang_ref
#define RECYCLER_NEXT list.head
#include <common/recycler.h>

#define STACK_TYPE char
#define STACK_PREFIX(X) id_##X
#include <common/stack.h>

struct context *context_top;

struct import_find
{
  const char *name;
  const char *fname;
};

static void import_find(struct strlist *path, struct import_find *data)
{
  if(data->fname)
    return;
  char *fname = strpool_concat_borrow(&strpool, path->string, "/");
  fname = strpool_concat_borrow(&strpool, fname, data->name);
  if(file_exists(fname))
    data->fname = strpool_concat(&strpool, fname, "");
}

static struct lang_node *src_get(const char *filename, bool toplevel, struct mempool_core *tmp_pool,
    struct id_stack *ids);

static struct lang_node *module_get(const char *name, struct mempool_core *tmp_pool, struct id_stack *ids)
{
  struct import_find import_find_data = {
    .name = name
  };
  strlist_walk(&config_import_path_list, &import_find_data, (strlist_walk_func) import_find);
  if(!import_find_data.fname)
    return NULL;
  struct lang_node *result = src_get(import_find_data.fname, false, tmp_pool, ids);
  return result;
}

static void imports_resolve(struct block_scan_result *block, struct mempool_core *tmp_pool, struct id_stack *ids)
{
  for(struct lang_ref *ref = block->sub_files.head; ref; ref = ref->list.head)
  {
    if(ref->node->type == ln_import)
    {
      assert(ref->node->children_head);
      dprint("Import %s", ref->node->children_head->data);
      if(!strcmp(ref->node->children_head->data, "BUILTIN"))
      {
        if(!block->global_context->cache_valid) // Newly created, not cached
          contexts_plug(block->global_context, context_builtin);
      }
      else
      {
        size_t size = ids->size;
        struct lang_node *module = module_get(ref->node->children_head->data, tmp_pool, ids);
        if(!module)
          die("%s:%u:Module %s not found", ref->node->children_head->filename, ref->node->children_head->line,
              ref->node->children_head->data);
        contexts_plug(block->global_context, module->block_scan->global_context);
        imports_resolve(module->block_scan, tmp_pool, ids);
        ids->size = size;
      }
    }
  }
  for(struct lang_ref *ref = block->sub_files.head; ref; ref = ref->list.head)
  {
    if(ref->node->type != ln_import)
    {
      assert(ref->node->type == ln_descend);
      struct variable *files = expr_evaluate(block->global_context, ref->node, tmp_pool, true);
      for(size_t i = 0; i < files->count; i ++)
      {
        assert(files->values[i].type == vt_string);
        const char *name = files->values[i].string;
        if(name[strlen(name) - 1] == '/') // It's a dir, append the REM
          name = strpool_concat(&strpool, name, conf_filename);
        if(!file_exists(name))
          die("Script file %s does not exist", name);
        dprint("Descending file %s", name);
        size_t size = ids->size;
        struct lang_node *node = src_get(name, false, tmp_pool, ids);
        assert(node->block_scan->global_context);
        contexts_plug(block->global_context, node->block_scan->global_context);
        imports_resolve(node->block_scan, tmp_pool, ids);
        ids->size = size;
      }
    }
  }
  for(struct lang_ref *ref = block->sub_blocks.head; ref; ref = ref->list.head)
    if(ref->node->block_scan && ref->node->block_scan->global_context)
      imports_resolve(ref->node->block_scan, tmp_pool, ids);
}

static struct sigaction oldact;

static void restart_prepare(void)
{
  if(!config_hierarchy_rebuild)
    if(sigaction(SIGSEGV, &oldact, NULL))
      die("Could not restore signal (%s)", strerror(errno));
}

static struct lang_node *src_get(const char *filename, bool toplevel, struct mempool_core *tmp_pool,
    struct id_stack *ids)
{
  for(const char *c = filename; *c; c ++)
    id_push(ids, *c);
  id_push(ids, '\0');
  struct lang_node *tree;
  if(!cache_ready || cached_file_changed(filename) || !(tree = cache_lang_tree_get(ids->data, toplevel)))
  {  // Don't be lazy and parse it
    dprint("Source tree %s not found in cache", filename);
    if(cache_ready && !config_hierarchy_rebuild)  // TODO better way to handle one source file change?
    {
      restart_prepare();
      config_hierarchy_rebuild = true;
      id_free(ids);
      restart_use(NULL);
    }
    tree = parse_file(filename);
    block_scan(tree, toplevel);
    cache_lang_tree_push(tree, ids->data);
    cache_file_mark(filename);
  }
  if(tree->block_scan->parent_file && toplevel)
  {
    struct variable *parent = expr_evaluate(tree->block_scan->global_context, tree->block_scan->parent_file, tmp_pool,
        true);
    if(parent->count != 1)
      die("Parent of %s must have exactly 1 value", filename);
    if(parent->values[0].type != vt_string)
      die("Parent of %s must be string", filename);
    id_free(ids);
    restart_prepare();
    restart_use(parent->values[0].string);
  }
  id_pop(ids);
  id_push(ids, '\n');
  return tree;
}

static void ref_push(struct lang_ref_list *list, struct lang_node *node, struct ref_rec_t *rec)
{
  struct lang_ref *ref = ref_rec_get(rec);
  ref->node = node;
  lang_ref_list_prepend(list, ref);
}

static void cache_tree(struct lang_ref *ref, struct ref_rec_t *rec)
{
  cache_lang_tree_push(ref->node, ref->node->filename);
  ref_rec_put(rec, ref);
}

static jmp_buf hier_jmp;

static void hierarchy_save(int unused)
{
  longjmp(hier_jmp, 1);
}

void hierarchy_build(void)
{
  struct mempool_core mempool;
  mempool_core_init(&mempool, 0, true);
  // TODO Is the cache needed in future? Maybe in evals?
  cache_load(cache_filename);
  context_toplevel_load(&mempool);
  struct id_stack ids;
  id_init(&ids);
  /*
   * FIXME: There is a bug in loading from cache. However, the bug is more a design fault than wrong code,
   *   so fixing it is probably impossible.
   *
   *   This at last eliminates the bug in an really ugly way. When the rewrite of cache & company comes at hand,
   *   this needs to be removed.
   */
  struct sigaction oldact, newact = {
    .sa_handler = hierarchy_save,
    .sa_flags = SA_NODEFER | SA_RESETHAND
  };
  if(!config_hierarchy_rebuild)
    if(sigaction(SIGSEGV, &newact, &oldact))
      die("Could not set signal (%s)", strerror(errno));
  if(setjmp(hier_jmp))
  {
    fprintf(stderr, "Emergency bug recover used, please fix the code, catching SEGFAULT is ugly\n");
    restart_prepare();
    config_hierarchy_rebuild = true;
    id_free(&ids);
    restart_use(NULL);
  }
  struct lang_node *current = src_get(conf_filename_start, true, &mempool, &ids);
  imports_resolve(current->block_scan, &mempool, &ids);
  if(!config_hierarchy_rebuild)
    if(sigaction(SIGSEGV, &oldact, NULL))
      die("Could not restore signal (%s)", strerror(errno));
  context_top = current->block_scan->global_context;
  id_free(&ids);
  mempool_core_clean(&mempool);
}
