/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "coroutine.h"
#include "external.h"
#include "flags.h"

#include <common/debug.h>
#include <common/mempool_core.h>

#include <string.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <stddef.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/wait.h>

static int children_ready, children_pipe;
static size_t avail_slots = 3; // TODO configure
static bool avail_count = true;

struct external_cmd
{
  pid_t pid;
  int pipe;
  bool terminated:1;
  bool has_output:1;
  struct mempool_core *mempool;
  char *output;
  size_t output_pos, output_size;
  int result;
  struct external_cmd *next, *prev;
};

struct external_cmd_list
{
  struct external_cmd *head, *tail;
};

#define LIST_PREFIX(X) cmd_##X
#define LIST_NODE struct external_cmd
#define LIST_CONTAINER struct external_cmd_list
#define LIST_NEXT next
#define LIST_PREV prev
#define LIST_TAIL tail
#define LIST_HEAD head
#include <common/list.h>

static struct external_cmd_list active;

static void child_died(int unused)
{
  if(write(children_pipe, " ", 1) == -1 && errno != EAGAIN)
    die("Could not write to self-pipe (%s)", strerror(errno));
}

void coroutine_init(void)
{
  int pipes[2];
  if(pipe(pipes))
    die("Could not create pipe (%s)", strerror(errno));
  fd_nonblock(pipes[0]);
  fd_nonblock(pipes[1]);
  children_ready = pipes[0];
  children_pipe = pipes[1];
  if(sigaction(SIGCHLD, &(struct sigaction) {
      .sa_handler = child_died,
      .sa_flags = SA_NOCLDSTOP
    }, NULL))
    die("Could not set signal handler for SIGCHLD (%s)",
        strerror(errno));
}

static void cmd_close(struct external_cmd *cmd, void *unused)
{
  if(!cmd->has_output && close(cmd->pipe))
    die("Could not close pipe to child %u/%d (%s)", (unsigned) cmd->pid,
        cmd->pipe, strerror(errno));
}

void coroutine_preexec(void)
{
  if(close(children_ready) || close(children_pipe))
    die("Could not close self-pipe (%s)", strerror(errno));
  cmd_walk(&active, NULL, cmd_close);
}

void coroutine_start(coroutine_func func, void *data)
{
  dprinte("Coroutine start");
#ifdef CAN_UCONTEXT
#error Not implemented
#else
  func(data);
#endif
}

struct fds_scan
{
  int *max_fds;
  fd_set *reads;
};

static void fds_scan(struct external_cmd *cmd, struct fds_scan *scan)
{
  if(cmd->has_output)
    return;
  FD_SET(cmd->pipe, scan->reads);
  if(*scan->max_fds < cmd->pipe)
    *scan->max_fds = cmd->pipe;
}

struct process_find
{
  pid_t pid;
  int pipe;
  struct external_cmd *cmd;
};

static void process_find(struct external_cmd *cmd,
    struct process_find *data)
{
  if(cmd->pid == data->pid || cmd->pipe == data->pipe)
    data->cmd = cmd;
}

static bool cmd_done(struct external_cmd *cmd)
{
  if(cmd->has_output && cmd->terminated)
  {
#ifdef CAN_UCONTEXT
#error Not implemented
#endif
    cmd_remove(&active, cmd);
    if(avail_count)
      avail_slots ++;
    return true;
  }
  return false;
}

struct output_read
{
  fd_set *fds;
  bool *changed;
};

static void output_read(struct external_cmd *cmd,
    struct output_read *data)
{
  if(cmd->has_output || !FD_ISSET(cmd->pipe, data->fds))
    return;
  dprint("Data on child process %u/%d", (unsigned) cmd->pid, cmd->pipe);
  if(!cmd->output)
    cmd->output = mempool_alloc_resizeable(cmd->mempool,
        &cmd->output_size);
  ssize_t amount;
  do
  {
    if(cmd->output_pos == cmd->output_size)
      cmd->output = mempool_grow(cmd->mempool, cmd->output,
          &cmd->output_size);
    if(!(amount = read(cmd->pipe, cmd->output + cmd->output_pos,
        cmd->output_size - cmd->output_pos - 1)))
    {
      dprinte("End of child output");
      cmd->has_output = true;
      cmd->output[cmd->output_pos] = '\0';
      cmd->output = mempool_size_set(cmd->mempool, cmd->output,
          cmd->output_pos + 1);
      *data->changed = cmd_done(cmd) || *data->changed;
      if(close(cmd->pipe))
        die("Can not close pipe to external command (%s)",
            strerror(errno));
      return;
    }
    cmd->output_pos += amount;
  } while(amount > 0);
  if(errno != EAGAIN)
    die("Error reading from external command (%s)", strerror(errno));
}

void coroutine_scheduler(bool block_for_child)
{
  if(!active.head)
    return; // Nothing to schedule
  bool children_changed = false;
  do
  {
    dprint("Coroutine scheduler tick (block=%d)", block_for_child);
    int max_fds = children_ready;
    fd_set reads;
    FD_ZERO(&reads);
    FD_SET(children_ready, &reads);
    cmd_walk(&active, &(struct fds_scan) {
        .max_fds = &max_fds,
        .reads = &reads
      }, (cmd_walk_func) fds_scan);
    if(select(max_fds + 1, &reads, NULL, NULL,
          block_for_child ? NULL : &(struct timeval) {}) == -1)
    {
      if(errno == EINTR)
        continue;
      die("Could not select (%s)", strerror(errno));
    }
    dprinte("Scheduler tick finished");
    // Some child died?
    if(FD_ISSET(children_ready, &reads))
    {
      // Read the whole pipe
      char buffer[1024];
      if(read(children_ready, buffer, 1024) == -1 && errno != EAGAIN)
        die("Could not read from self-pipe (%s)", strerror(errno));
      int status;
      pid_t child;
      while(child = waitpid(-1, &status, WNOHANG))
      {
        if(child == -1)
        {
          if(errno == ECHILD)
            break;  // No more children
          die("Error waiting for child (%s)", strerror(errno));
        }
        struct process_find pf = {
          .pid = child
        };
        cmd_walk(&active, &pf, (cmd_walk_func) process_find);
        dprint("Child %u/%d terminated", (unsigned) pf.cmd->pid,
            pf.cmd->pipe);
        pf.cmd->result = status;
        pf.cmd->terminated = true;
        children_changed = cmd_done(pf.cmd) || children_changed;
      }
    }
    cmd_walk(&active, &(struct output_read) {
        .fds = &reads,
        .changed = &children_changed
      }, (cmd_walk_func) output_read);
  }
  while(block_for_child && !children_changed);
}

char *coroutine_run_command(const char *command[], bool capture, struct mempool_core *mempool, int *result,
    const char *workdir)
{
  if(avail_count)
  {
    while(!avail_slots)
      coroutine_scheduler(true);  // Too many running
    avail_slots --;
  }
  int cmdpipe = 0;
  struct external_cmd cmd = {
    .pid = external_run(command, capture ? &cmdpipe : NULL, workdir),
    .has_output = !capture,
    .mempool = mempool
  };
  cmd.pipe = cmdpipe;
  cmd_append(&active, &cmd);
#ifdef CAN_UCONTEXT
#error Not implemented
#else
  coroutine_scheduler(true); // This is the only command, wait for it
  *result = cmd.result;
  return cmd.output;
#endif
}
