/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "block.h"
#include "node.h"
#include "context.h"

#include <common/debug.h>
#include <common/config.h>

#include <stdbool.h>
#include <stdio.h>

static void block_scan_init(struct block_scan_result *block)
{
  identifier_list_drop(&block->mentioned);
  identifier_list_drop(&block->out);
  lang_ref_list_drop(&block->sub_files);
  lang_ref_list_drop(&block->sub_blocks);
  block->parent_file = NULL;
}

#define MEMPOOL_PREFIX(X) block_scan_pool_##X
#define MEMPOOL_TYPE struct block_scan_result
#define MEMPOOL_INIT block_scan_init
#define MEMPOOL_HARDCODE_VARIABLE &persistent_pool
#include <common/mempool.h>

struct flags {
  bool started:1;       // Used to descend only the top-level block
  bool local_compute:1; // Identifier must be computed locally to this block
  bool block_entry:1;   // Entry-point to sub-block
  bool file_entry:1;    // Entry to another file
  // Direct mapping to variable flags from the language
  bool var_inherited:1;
  bool var_export:1;
  bool var_global:1;
  bool var_nocache:1;

  bool expect_lvalue:1;     // This must be an L-value
  bool keep_context:1;      // Leave the context alone? Default is eat it
  bool keep_assign_flags:1;
  bool cant_subscript:1;    // Inside a subscribed item already
  bool create_ident:1;      // This creates an identifier, if it does not exist
  bool direct_assign:1;     // The assignment is direct
  bool is_identifier:1;
  bool is_eval:1;
  bool is_multi_instance:1;
  bool is_map:1;
  bool in_class:1;
};

struct event                // Single event, found in the tree, to be processed
{
  struct lang_node *node;   // Which node triggers the event?
  struct lang_node *context;// Where does the node belong?
  struct flags flags;       // Some flags, like if it is left side of
                            // assignment, if it should be global, etc
  struct event *next;
};

struct scan_data
{
  struct identifier_list ignore_read;
  struct event *head, *tail;
  struct lang_node *current_context;
  struct flags flags;
  struct block_scan_result *result;
};

#define LIST_PREFIX(X) scan_##X
#define LIST_CONTAINER struct scan_data
#define LIST_NODE struct event
#define LIST_NEXT next
#define LIST_HEAD head
#define LIST_TAIL tail
#include <common/list.h>

static struct mempool_core tmp_pool;

#define MEMPOOL_PREFIX(X) event_pool_##X
#define MEMPOOL_TYPE struct event
#define MEMPOOL_HARDCODE_VARIABLE &tmp_pool
#include <common/mempool.h>

static size_t scan_cnt;

static void event_new(struct lang_node *node, struct flags flags, struct lang_node *context, struct scan_data *data)
{
  dprint("New event on %u:%s ctx: %p l:%d be: %d ci: %d", node->line,
      node->data, (void *) context, (int) flags.local_compute,
      (int) flags.block_entry, (int) flags.create_ident);
  struct event *event = event_pool_get();
  event->node = node;
  event->context = context;
  event->flags = flags;
  scan_append(data, event);
}

struct extract_scan
{
  struct lang_node *context;
  struct scan_data *data;
};

static void node_scan(struct lang_node *node, struct scan_data *data);

static void extract_scan(struct lang_node *node, struct extract_scan *data)
{
  if(node->next)  // Not the last one
  {
    // TODO check if it is the right kind of symbol
    event_new(node, (struct flags) {
        .create_ident = true,
        .is_identifier = true,
        .direct_assign = true
      }, data->context, data->data);
  }
  else
  {
    node_scan(node, data->data);
  }
}

struct def_scan
{
  struct lang_node *context;
  struct scan_data *data;
  bool is_class:1;
};

static void def_scan(struct lang_node *node, struct def_scan *data)
{
  switch(node->type)
  {
    case ln_param_multi:
    case ln_param_single:
    {
      if(identifier_find(&data->data->ignore_read, node->data))
        die("%s:%u:Multiple parameters with the same name %s", node->filename, node->line, node->data);
      struct identifier *id = identifier_pool_get(&tmp_pool);
      id->name = node->data;
      identifier_list_prepend(&data->data->ignore_read, id);
      break;
    }
    case ln_inherit_alter:
    case ln_inherit_direct:
      langlist_walk(node, data->data, (langlist_walk_func) node_scan);
      break;
    case ln_block:
      assert(!node->next);
      if(data->is_class)
        event_new(node, (struct flags) {
            .block_entry = true,
            .is_multi_instance = true
          }, data->context, data->data);
      break;
    default:
      die("%s:%u:Invalid syntax in class definition", node->filename, node->line);
  }
}

// Walk the tree and extract events
static void node_scan(struct lang_node *node, struct scan_data *data)
{
  assert(node->type == ln_block || data->flags.started);
  node->block_scan = data->result;
  struct flags flags = data->flags;// Store flags. This or the descending scans
                                   // may want to overwrite it, and it must be
                                   // returned to proper state before leaving
  bool descend = false;            // Descend to subnodes?
  struct flags local_flags = {};
  bool is_lvalue = false;          // Am I an L-Value? Depends on if it can be
                                   // and is expected
  bool clean_flags = false, clean_context = false;
  bool can_var_flags = false;
  switch(node->type)
  {
    case ln_none:                  // This should never get out
    case ln_param_multi:           // Handled elsewhere
    case ln_param_single:
    case ln_inherit_alter:
    case ln_inherit_direct:
      assert(0);
      break;
    case ln_string:                // Just a constant, nothing interesting
      break;                       // Contains no subnodes
    case ln_block:                 // Descend to the top-level one only
      descend = !data->flags.started;
      if(!descend)                 // Generate link to the next one
        event_new(node, (struct flags) { .block_entry = true }, NULL, data);
      data->flags.started = true;
      break;
    case ln_variable:      // Store the variable and store its context
    case ln_variable_escape:
      flags.is_identifier = true;
      if(flags.direct_assign || !identifier_find(&data->ignore_read, node->data))
        event_new(node, flags, data->current_context, data);
      clean_flags = !flags.keep_assign_flags;
      clean_context = !flags.keep_context;
      is_lvalue = true;    // Simple variable is L-Value no matter what
      can_var_flags = true;
      break;
    case ln_descend:      // Record them and mark anything inside it must be
                          // computable _now_
      local_flags.block_entry = true;
      // No break; -- fall trough
    case ln_parent:
      local_flags.file_entry = true;
      event_new(node, local_flags, NULL, data);
      descend = true;
      data->flags.local_compute = true;
      break;
    case ln_import:
      event_new(node, (struct flags) { .file_entry = true }, NULL, data);
      break;
    case ln_array_concat:
      if(flags.expect_lvalue)
      {
        is_lvalue = data->flags.keep_context = data->flags.keep_assign_flags = true;
        clean_flags = clean_context = true;
      }
      can_var_flags = true;
      // Fall trough
    case ln_shell:
    case ln_perl:
    case ln_string_concat:
      descend = true;
      break;
    case ln_function:
      event_new(node, flags, NULL, data);
      descend = true;
      break;
    case ln_assign_direct:
      data->flags.direct_assign = true;
      // Fall trough
    case ln_assign_alter:
      data->flags.expect_lvalue = true;
      data->flags.create_ident = true;
      data->current_context = node;
      descend = true;
      break;
    case ln_eval:
      event_new(node, (struct flags) {
          .block_entry = true, .is_eval = true,
          .is_map = true
        }, NULL, data);
      descend = true;
      break;
    case ln_map:
    {
      struct lang_node *block = langlist_last(node);
      if(block->type != ln_block)
        die("%s:%u:Syntax error in map", node->filename, node->line);
      event_new(block, (struct flags) {
          .block_entry = true,
          .is_multi_instance = true
        }, node, data);
      break;
    }
    case ln_extract:
      langlist_walk(node, &(struct extract_scan) {
        .context = node,
        .data = data
      }, (langlist_walk_func) extract_scan);
      break;;
    case ln_fun_def:
      event_new(node, (struct flags) {
          .create_ident = true,
          .direct_assign = true
        }, NULL, data);
      if(data->flags.in_class)
      {
        struct def_scan sdata = {
          .context = node,
          .data = data
        };
        langlist_walk(node, &sdata, (langlist_walk_func) def_scan);
        identifier_list_drop(&data->ignore_read);
      }
      break;
    case ln_class_def:
      event_new(node, (struct flags) {
          .create_ident = true,
          .direct_assign = true
        }, NULL, data);
      struct def_scan sdata = {
        .context = node,
        .data = data,
        .is_class = true
      };
      langlist_walk(node, &sdata, (langlist_walk_func) def_scan);
      identifier_list_drop(&data->ignore_read);
      break;
              // Some flags for variables
#define VAR_FLAG(v) case ln_##v: data->flags.var_##v = descend = \
      can_var_flags = true; if(data->flags.expect_lvalue) is_lvalue = clean_flags = \
      true; break
    VAR_FLAG(inherited);
    VAR_FLAG(export);
    VAR_FLAG(global);
    VAR_FLAG(nocache);
    case ln_method:
    case ln_subscript:
      diee("Not implemented");
    // TODO the rest
  }
  if(flags.expect_lvalue && !is_lvalue)
    die("%s:%u:Expected a l-value", node->filename, node->line);
  if(!can_var_flags && (flags.var_inherited || flags.var_export || flags.var_global || flags.var_nocache))
    die("%s:%u:Only variables can have flags or specifiers", node->filename, node->line);
  if(descend)
  {
    langlist_walk(node, data, (langlist_walk_func) node_scan);
    data->flags = flags;
  }
  if(clean_flags)
    data->flags.expect_lvalue = data->flags.direct_assign =
        data->flags.create_ident = false;
  if(clean_context)
    data->current_context = NULL;
}

static struct block_scan_result *block_scan_internal(struct lang_node *block, bool is_class, bool is_toplevel);

static void event_process_1(struct event *event, struct block_scan_result *result)
{
  if(event->flags.create_ident || event->flags.var_global ||
      event->flags.var_export)
  {
    if(event->flags.var_inherited)
      die("%s:%u:Assignment to inherited '%s'", event->node->filename, event->node->line, event->node->data);
    struct identifier *idf = identifier_find(&result->out,
        event->node->data);
    if(idf && idf->overwritten && event->flags.direct_assign)
      die("%s:%u:Direct assignments to '%s' which is already assigned to", event->node->filename, event->node->line,
          event->node->data);
    if(!idf)
    {
      idf = identifier_pool_get(&persistent_pool);
      idf->name = event->node->data;
      identifier_list_append(&result->out, idf);
    }
    idf->overwritten = event->flags.direct_assign;
    idf->global |= event->flags.var_global;
    idf->exported |= event->flags.var_export;
    if(event->flags.create_ident)
    {
      struct node_bind *node = identifier_node_pool_get();
      node->absolute = event->flags.direct_assign;
      node->node = event->context ? event->context : event->node;
      node_bind_append(&idf->creators, node);
    }
  }
  else if(event->flags.file_entry && event->node->type == ln_parent)
  {
    dprinte("Found parent directory");
    if(result->parent_file)
      die("%s:multiple parent files", event->node->filename);
    result->parent_file = event->node;
  }
  else if(event->flags.file_entry)
  {
    dprinte("Found sub file");
    lang_ref_prepend(&result->sub_files, event->node);
  }
  if(event->flags.block_entry && !event->flags.file_entry && !event->flags.is_eval)
  {
    dprinte("Found sub-block in the same file, scanning too");
    event->node->block_scan = NULL; // It was wrongly filled, remove it
    block_scan_internal(event->node, event->context && event->context->type == ln_class_def, false);
    dprinte("Sub-block scan finished");
    if(!event->flags.create_ident)  // Not a function nor a class
      lang_ref_prepend(&result->sub_blocks, event->node);
  }
}

static void identifier_print(struct identifier *idf, void *unused)
{
  putc(' ', stderr);
  fputs(idf->name, stderr);
  putc('\n', stderr);
}

static void event_process_2(struct event *event,
    struct block_scan_result *result)
{
  if(event->flags.is_identifier && !event->flags.var_inherited &&
      !event->flags.direct_assign)
  {
    struct identifier *idf = identifier_find(&result->mentioned,
        event->node->data);
    if(!idf)
    {
      idf = identifier_pool_get(&persistent_pool);
      idf->symlink = identifier_find(&result->out, event->node->data);
      idf->name = event->node->data;
      identifier_list_append(&result->mentioned, idf);
    }
    struct identifier *master = idf->symlink ? idf->symlink : idf;
    if(event->flags.local_compute)
    {
      if(!idf->symlink)
        die("%s:%u:local variable '%s' not defined locally",
            event->node->filename, event->node->line,
            event->node->data);
      master->local_compute = true;
    }
    master->nocache = master->nocache || event->flags.var_nocache;
  }
}

static void con_plug(struct lang_ref *ref, struct context *parent)
{
  struct block_scan_result *result = ref->node->block_scan;
  if(result->global_context)
    contexts_plug(parent, result->global_context);
  lang_ref_list_walk(&result->sub_blocks, result->global_context, (lang_ref_list_walk_func) con_plug);
}

/*
 * The task is to create some info about a block. It should contain:
 * - List of identifiers it exports
 * - List of identifiers it may import
 * - List of identifiers it needs (is it recognizable at this level?)
 * - List of some dynamic identifiers
 * - List of block it includes
 * - List of files it references
 */
static struct block_scan_result *block_scan_internal(struct lang_node *node, bool is_class, bool is_toplevel)
{
  // Already computed?
  if(node->block_scan)
    return node->block_scan;
  // Scan the tree and extract important events into a list
  struct block_scan_result *result = block_scan_pool_get();
  struct scan_data data = { .flags = { .in_class = is_class }, .result = result };
  scan_cnt ++;
  scan_drop(&data);
  node_scan(node, &data);
  identifier_list_drop(&data.ignore_read);
  // First phase ‒ get created identifiers
  scan_walk(&data, result, (scan_walk_func) event_process_1);
#ifdef TRACE
  if(config_debug)
  {
    fputs("List of created identifiers:\n", stderr);
    identifier_list_walk(&result->out, NULL, identifier_print);
  }
#endif
  scan_walk(&data, result, (scan_walk_func) event_process_2);
#ifdef TRACE
  if(config_debug)
  {
    fputs("List of identifiers the block talks about\n", stderr);
    identifier_list_walk(&result->mentioned, NULL, identifier_print);
  }
#endif
  // Cleanup
  if(!-- scan_cnt)
  {
    dprinte("Cleaning scan temporary data");
    mempool_core_clean(&tmp_pool);
  }
  result->node = node;
  if(!is_class)
  {
    context_raw_block(result, node);
    if(is_toplevel)
      contexts_plug(context_toplevel, result->global_context);
    lang_ref_list_walk(&result->sub_blocks, result->global_context, (lang_ref_list_walk_func) con_plug);
  }
  return node->block_scan = result;
}

struct block_scan_result *block_scan(struct lang_node *node, bool toplevel)
{
  return block_scan_internal(node, false, toplevel);
}

struct block_scan_result *block_scan_class(struct lang_node *node)
{
  return block_scan_internal(node, true, false);
}

void lang_block_init(void)
{
  mempool_core_init(&tmp_pool, 0, true);
}

void lang_block_restart(void)
{
  scan_cnt = 0;
}
