/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "builtin.h"
#include "context.h"

#include <common/debug.h>
#include <common/stringpool.h>

#include <stdio.h>
#include <string.h>

static struct var_dep_list *param_deps_merge(struct variable **params, size_t count)
{
  struct var_dep_list *result = NULL;
  for(size_t i = 0; i < count; i ++)
    result = var_dep_lists_merge(result, params[i]->deps);
  return result;
}

/*
 * Params: string to eat from, set of characters after to stop.
 */
static struct variable *eat_first(struct variable **params, size_t count, struct context *context,
    struct context *parent, struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node,
    struct function *function, bool object_toplevel)
{
  if(params[0]->values[0].type != vt_string && params[0]->values[0].type != vt_object)
    die("%s:%u:Can not eat in something not a string or object", lang_node->filename, lang_node->line);
  if(params[1]->values[0].type != vt_string)
    die("%s:%u:Can not eat something not a string", lang_node->filename, lang_node->line);
  size_t len = strlen(params[0]->values[0].string);
  size_t start = strcspn(params[0]->values[0].string, params[1]->values[0].string);
  struct variable *result = var_alloc(tmp_pool, 1, temporary);
  result->values[0] = (struct var_value) {
    .type = vt_string,
    .str_temporary = params[0]->values[0].str_temporary,
    .string = params[0]->values[0].string + start + (start != len)
  };
  if(result->values[0].str_temporary && !temporary)
  {
    result->values[0].string = strpool_dup(&strpool, result->values[0].string);
    result->values[0].str_temporary = false;
  }
  result->deps = param_deps_merge(params, count);
  result->state = vs_computed;
  return result;
}

#define STACK_PREFIX(X) strbuf_##X
#define STACK_TYPE char *
#include <common/stack.h>

/*
 * Params: string to split, list of non-delimiters (like '\\', '\ '), list of delimiters
 */
static struct variable *split_escaped(struct variable **params, size_t count, struct context *context,
    struct context *parent, struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node,
    struct function *function, bool object_toplevel)
{
  if(params[0]->values[0].type != vt_string && params[0]->values[0].type != vt_object)
    die("%s:%u:Can not split something not a string or object", lang_node->filename, lang_node->line);
  for(size_t i = 0; i < params[1]->count; i ++)
    if(params[1]->values[i].type != vt_string)
      die("%s:%u:Non-delimiter must be a string", lang_node->filename, lang_node->line);
  for(size_t i = 0; i < params[2]->count; i ++)
    if(params[2]->values[i].type != vt_string)
      die("%s:%u:Delimiter must be a string", lang_node->filename, lang_node->line);
  bool in_delim = true;
  struct strbuf_stack strbuf;
  strbuf_init(&strbuf);
  const char *str = params[0]->values[0].string, *last;
  struct mempool_core *strpool_local = temporary ? tmp_pool : &strpool;
  while(*str)
  {
    DELIM:
    for(size_t i = 0; i < params[1]->count; i ++)
    {
      size_t len = strlen(params[1]->values[i].string);
      if(!strncmp(params[1]->values[i].string, str, len))
      {
        if(in_delim)
          last = str;
        str += len;
        goto NO_DELIMIT;
      }
    }
    for(size_t i = 0; i < params[2]->count; i ++)
    {
      size_t len = strlen(params[2]->values[i].string);
      if(!strncmp(params[2]->values[i].string, str, len))
      {
        if(!in_delim)
        {
          in_delim = true;
          strbuf_push(&strbuf, strpool_dupn(strpool_local, last, str - last));
        }
        str += len;
        goto DELIM; // Continue, but to the topmost loop
      }
    }
    if(in_delim)
      last = str;
    str ++;
    NO_DELIMIT:
    if(in_delim)
      in_delim = false;
  }
  if(!in_delim)
    strbuf_push(&strbuf, strpool_dupn(strpool_local, last, str - last));
  struct variable *result = var_alloc(tmp_pool, strbuf.size, temporary);
  for(size_t i = 0; i < strbuf.size; i ++)
    result->values[i] = (struct var_value) {
      .type = vt_string,
      .string = strbuf.data[i],
      .str_temporary = temporary
    };
  strbuf_free(&strbuf);
  result->deps = param_deps_merge(params, count);
  return result;
}

static struct variable *echo(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  assert(count == 1);
  assert(params[0]->count = 1);
  if(params[0]->values[0].type != vt_string && params[0]->values[0].type != vt_object)
    die("%s:%u:Can not echo something not an object or string", lang_node->filename, lang_node->line);
  puts(params[0]->values[0].string);
  return &var_empty;
}

static struct variable *iffun(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  struct variable *which = params[0]->count ? params[1] : params[2];
  size_t size = which->count;
  struct variable *result = var_alloc(tmp_pool, size, temporary);
  memcpy(result, which, sizeof *result + size * sizeof *result->values);
  result->temporary = temporary;
  if(!temporary)
    var_values_make_persistent(result);
  result->deps = var_dep_lists_merge(params[0]->deps, which->deps);
  return result;
}

static struct variable *defined(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  if(params[0]->undefined)
    return params[0];
  struct variable *result = var_alloc(tmp_pool, 1, temporary);
  result->values[0] = (struct var_value) {
    .type = vt_string,
    .string = "defined"
  };
  result->deps = params[0]->deps;
  return result;
}

static struct variable *loc(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  if(params[0]->values[0].type != vt_string)
    die("%s:%u:Parameter of loc must be a string", lang_node->filename, lang_node->line);
  const char *slash = rindex(context->filename, '/');
  if(slash)
  {
    size_t dlen = slash - context->filename + 1, flen = strlen(params[0]->values[0].string);
    struct variable *result = var_alloc(tmp_pool, 1, temporary);
    char *str = mempool_alloc(temporary ? tmp_pool : &strpool, dlen + flen + 1);
    *result->values = (struct var_value) {
      .type = vt_string,
      .str_temporary = temporary,
      .string = str
    };
    memcpy(str, context->filename, dlen);
    memcpy(str + dlen, params[0]->values[0].string, flen);
    str[dlen + flen] = '\0';
    result->deps = params[0]->deps;
    return result;
  }
  else
  {
    if(temporary)
      return params[0];
    else
      return var_make_persistent(params[0]);
  }
}

static struct variable *env(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  if(params[0]->values[0].type != vt_string)
    die("%s:%u:Parameter of env must be a string", lang_node->filename, lang_node->line);
  const char *res_str = strpool_dup(pool_choose(tmp_pool, temporary), getenv(params[0]->values[0].string));
  struct variable *result = var_alloc(tmp_pool, !!res_str, temporary);
  result->deps = params[0]->deps;
  result->state = vs_computed;
  if(res_str)
    *result->values = (struct var_value) {
      .type = vt_string,
      .str_temporary = temporary,
      .string = res_str
    };
  return result;
}

static struct variable *concat(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  size_t output_size = 0, output_sizes[params[0]->count];
  for(size_t i = 0; i < params[0]->count; i ++)
  {
    if(params[0]->values[i].type != vt_string && params[0]->values[i].type != vt_object)
      die("%s:%u:Can not concat functions or classes", lang_node->filename, lang_node->line);
    output_size += output_sizes[i] = strlen(params[0]->values[i].string);
  }
  struct variable *result = var_alloc(tmp_pool, 1, temporary);
  result->deps = params[0]->deps;
  result->state = vs_computed;
  char *res_str = mempool_alloc_nozero(temporary ? tmp_pool : &strpool, output_size + 1);
  res_str[output_size] = '\0';
  *result->values = (struct var_value) {
    .type = vt_string,
    .str_temporary = temporary,
    .string = res_str
  };
  for(size_t i = 0; i < params[0]->count; i ++)
  {
    memcpy(res_str, params[0]->values[i].string, output_sizes[i]);
    res_str += output_sizes[i];
  }
  return result;
}

static struct variable *boolean(bool value, bool temporary, struct mempool_core *tmp_pool, struct var_dep_list *deps)
{
  struct variable *result = var_alloc(tmp_pool, value, temporary);
  result->state = vs_computed;
  result->deps = deps;
  if(value)
  {
    *result->values = (struct var_value) {
      .type = vt_string,
      .str_temporary = false,
      .string = "true"
    };
  }
  return result;
}

static struct variable *eq(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  return boolean(var_eq(params[0], params[1]), temporary, tmp_pool, param_deps_merge(params, 2));
}

static struct variable *and(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  return boolean(params[0]->count && params[1]->count, temporary, tmp_pool, param_deps_merge(params, 2));
}

static struct variable *not(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  return boolean(!params[0]->count, temporary, tmp_pool, params[0]->deps);
}

static struct variable *vartype(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  struct variable *result = var_alloc(tmp_pool, 1, temporary);
  result->state = vs_computed;
  const char *str;
  switch(params[0]->values[0].type)
  {
    case vt_string:
      str = "string";
      break;
    case vt_object:
      str = "object";
      break;
    case vt_function:
      str = "function";
      break;
    case vt_class:
      str = "class";
      break;
    case vt_method:
      str = "method";
      break;
  }
  *result->values = (struct var_value) {
    .type = vt_string,
    .string = str
  };
  return result;
}

static struct variable *reverse(struct variable **params, size_t count, struct context *context, struct context *parent,
    struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node, struct function *function,
    bool object_toplevel)
{
  if(params[0]->values[0].type != vt_string && params[0]->values[0].type != vt_object)
    die("%s:%u:Can not reverse something not a string or object", lang_node->filename, lang_node->line);
  struct variable *result = var_alloc(tmp_pool, 1, temporary);
  result->state = vs_computed;
  size_t len = strlen(params[0]->values[0].string);
  char *res_str = mempool_alloc(temporary ? tmp_pool : &strpool, len + 1);
  *result->values = (struct var_value) {
    .type = vt_string,
    .string = res_str,
    .str_temporary = temporary
  };
  res_str[len] = '\0';
  for(size_t i = 0; i < len; i ++)
    res_str[len - 1 - i] = params[0]->values[0].string[i];
  return result;
}

static enum param_type singles[] = { pt_single, pt_single };
static enum param_type pars_s_m_m[] = { pt_single, pt_multi, pt_multi };
static enum param_type multies[] = { pt_multi, pt_multi, pt_multi };

#define F(NAME, PCOUNT, TYPES) { .name = #NAME, .code = NAME, .pcount = PCOUNT, .types = TYPES }
#define SF(NAME) F(NAME, 1, singles)
#define SF2(NAME) F(NAME, 2, singles)
#define MF(NAME) F(NAME, 1, multies)
#define MF2(NAME) F(NAME, 2, multies)

struct builtin_function builtin_functions[] = {
  SF(echo),
  SF2(eat_first),
  F(split_escaped, 3, pars_s_m_m),
  { .name = "if", .code = iffun, .pcount = 3, .types = multies },
  MF(defined),
  SF(loc),
  SF(env),
  MF(concat),
  MF2(eq),
  MF2(and),
  MF(not),
  SF(vartype),
  SF(reverse),
  { NULL, NULL, 0, NULL }
  // Append new functions to end, or you break cache
};

struct context *context_builtin;

void lang_builtin_init(void)
{
  context_builtin = context_get(&persistent_pool);
  context_builtin->builtin = true;
  for(size_t i = 0; builtin_functions[i].name; i ++)
  {
    dprint("Preparing builtin %s", builtin_functions[i].name);
    struct variable *var = var_alloc(NULL, 1, false);
    var->state = vs_computed;
    size_t pcount = 0;
    if(builtin_functions[i].pcount == -1)
      while(builtin_functions[i].types[pcount ++] != pt_stop);
    else
      pcount = builtin_functions[i].pcount;
    struct function *function = mempool_alloc(&persistent_pool, sizeof *function + pcount * sizeof *function->params);
    *function = (struct function) {
      .pcount = builtin_functions[i].pcount,
      .code = builtin_functions[i].code,
      .compiled = var
    };
    for(size_t j = 0; j < pcount; j ++)
      function->params[j].type = builtin_functions[i].types[j];
    builtin_functions[i].value = var->values[0] = (struct var_value) {
      .type = vt_function,
      .function = function
    };
    struct var_holder *holder = context_var_set(context_builtin, builtin_functions[i].name, var, false, false);
    holder->exported = true;
    struct context_plug *plug = mempool_alloc(&persistent_pool, sizeof *plug);
    plug->context = context_builtin;
  }
}
