/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "class.h"
#include "var.h"
#include "context.h"
#include "compute.h"
#include "object.h"

#include <common/mempool_core.h>
#include <common/debug.h>
#include <cache/context.h>

#include <assert.h>

struct inherit_plug
{
  struct context *object;
  struct context *parent;
  struct mempool_core *tmp_pool;
};

struct deps_cpy
{
  struct context *child;
  struct var_holder *target;
};

static void deps_cpy(struct context_plug *plug, struct deps_cpy *data)
{
  if(data->child && data->child == plug->context->parent_of)
    plug->context->dep_list = data->child->dep_list;
  if(!data->child || data->child == plug->context->parent_of)
  {
    context_var_reference(plug->context, "TARGET", data->target, true);
    plug_list_walk(&plug->context->parents, &(struct deps_cpy) { .child = plug->context, .target = data->target },
        (plug_list_walk_func) deps_cpy);
  }
}

void inherit_plug(struct lang_node *node, struct inherit_plug *data)
{
  struct variable *objects = expr_evaluate_ext(data->object, data->parent, node, data->tmp_pool, true, true, false);
  for(size_t i = 0; i < objects->count; i ++)
  {
    if(objects->values[i].type != vt_object)
      die("%s: %u: Trying to inherint something not an object (%s)", node->filename, node->line, node->data);
    contexts_plug(objects->values[i].object->context, data->object);
    objects->values[i].object->context->parent_of = data->object;
  }
  data->object->dep_list = var_dep_lists_merge(data->object->dep_list, objects->deps);
}

struct variable *object_build(struct variable **params, size_t param_count, struct context *parent,
    struct context *parent_context, struct mempool_core *tmp_pool, bool temporary, struct lang_node *node,
    struct function *function, bool object_toplevel)
{
  struct context *context = context_get(&persistent_pool);
  context->block = function->block->block_scan;
  context_prepare(context);
  struct var_dep_list *deps = function->compiled->deps;
  for(size_t i = 0; i < param_count; i ++)
  {
    context_var_set(context, function->params[i].name, var_make_persistent(params[i]), false, false);
    deps = var_dep_lists_merge(deps, params[i]->deps);
  }
  contexts_plug(function->func_context, context);
  assert(!parent_context->function);// Ups. TODO: make this function context permanent. Crazy idea, isn't it?
  // Or plug it's parent?
  if(parent_context != function->func_context)
    contexts_plug(parent_context, context);
  struct lang_node *parent_node = function->note;
  struct inherit_plug inherit_data = {
    .object = context,
    .tmp_pool = tmp_pool,
    .parent = parent_context
  };
  struct object *object = mempool_alloc(&persistent_pool, sizeof *object);
  *object = (struct object) {
    .context = context,
    .class_code = function->func
  };
  context->object = object;
  while(parent_node && parent_node->type != ln_block) // FIXME if there is :~, can't be anything else, etc
  {
    if(parent_node->type == ln_inherit_alter && !strcmp(parent_node->data, "-~"))
      assert(0);  // TODO
    else if(parent_node->type == ln_inherit_alter || parent_node->type == ln_inherit_direct)
      langlist_walk(parent_node, &inherit_data, (langlist_walk_func) inherit_plug);
    else
      assert(0);  // Can Not Happen
    parent_node = parent_node->next;
  }
  struct variable *target = variable_compute(context, "TARGET", tmp_pool);
  if(target->undefined)
    die("%s:%u:TARGET of the %s class is not defined", node->filename, node->line, function->func->data);
  struct variable *result = var_alloc(tmp_pool, target->count + !target->count, temporary);
  deps = var_dep_lists_merge(deps, target->deps);
  if(target->count)
    for(size_t i = 0; i < target->count; i ++)
    {
      assert(!target->values[i].str_temporary);
      if(target->values[i].type != vt_string && target->values[i].type != vt_object)
        die("%s:%u:target of the %s class is not string or object", node->filename, node->line, function->func->data);
      result->values[i] = (struct var_value) {
        .type = vt_object,
        .string = target->values[i].string,
        .object = target->values[i].type == vt_object ? target->values[i].object : object
      };
    }
  else
  {
    *result->values = (struct var_value) {
      .type = vt_object,
      .string = "",
      .object = object
    };
  }
  context->dep_list = result->deps = deps;
  deps_cpy(&(struct context_plug) { .context = context },
      &(struct deps_cpy) { .target = context_var_search(context, "TARGET", true) });
  context->cache_refs = result->count;
  cache_context_put(context, tmp_pool);
  return result;
}

struct class_def_scan
{
  size_t param_count, parent_count;
  struct lang_node *block, *parents_first;
};

static void class_def_scan(struct lang_node *node, struct class_def_scan* data)
{
  switch(node->type)
  {
    case ln_param_multi:
    case ln_param_single:
      data->param_count ++;
      break;
    case ln_inherit_alter:
    case ln_inherit_direct:
      if(!data->parent_count ++)
        data->parents_first = node;
      break;
    case ln_block:
      assert(!data->block);
      data->block = node;
      break;
    default:
      assert(0);
  }
}

struct class_fill
{
  struct param_def *param;
};

static void class_fill(struct lang_node *node, struct class_fill *data)
{
  switch(node->type)
  {
    case ln_param_multi:
    case ln_param_single:
      *data->param ++ = (struct param_def) {
        .name = node->data,
        .type = node->type == ln_param_multi ? pt_multi : pt_single
      };
      break;
    case ln_inherit_alter:
    case ln_inherit_direct:
      // TODO?
      break;
    case ln_block:
      break;
    default:
      assert(0);
  }
}

void class_build(struct lang_node *expr, struct context *context, struct mempool_core *tmp_pool, bool temporary,
    struct var_value *result, struct variable *compiled)
{
  struct class_def_scan def_scan = {};
  langlist_walk(expr, &def_scan, (langlist_walk_func) class_def_scan);
  *result = (struct var_value) {
    .type = vt_class,
    .function = mempool_alloc(pool_choose(tmp_pool, temporary), sizeof *result->function +
        def_scan.param_count * sizeof *result->function->params),
    .fun_temporary = temporary
  };
  *result->function = (struct function) {
    .pcount = def_scan.param_count,
    .code = object_build,
    .block = def_scan.block,
    .compiled = compiled,
    .func = expr,
    .note = def_scan.parents_first,
    .func_context = context
  };
  langlist_walk(expr, &(struct class_fill) {
      .param = result->function->params
    }, (langlist_walk_func) class_fill);
}
