/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "compute.h"
#include "node.h"
#include "var.h"
#include "context.h"
#include "class.h"
#include "function.h"
#include "block.h"

#include <common/mempool_core.h>
#include <common/debug.h>
#include <common/stringpool.h>
#include <interact/shell.h>
#include <cache/context.h>

#include <assert.h>
#include <string.h>
#include <stdio.h>

struct subs_eval
{
  struct variable **subs;
  struct context *context;
  struct mempool_core *pool;
  size_t val_count;
  size_t pos;
  struct var_dep_list *deps;
};

static void subs_eval(struct lang_node *expr, struct subs_eval *subs)
{
  subs->subs[subs->pos] = expr_evaluate(subs->context, expr, subs->pool, true);
  subs->deps = var_dep_lists_merge(subs->deps, subs->subs[subs->pos]->deps);
  subs->val_count += subs->subs[subs->pos ++]->count;
}

static struct variable *done(struct variable *var)
{
  var->state = vs_computed;
  return var;
}

static struct variable *string_concat_fun(struct variable **params, size_t count, struct context *context,
    struct context *parent, struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node,
    struct function *function, bool object_toplevel)
{
  struct mempool_core *pool = pool_choose(tmp_pool, temporary);
  // Compute length of the result
  size_t len = 0;
  size_t lens[count];
  struct var_dep_list *deps = NULL;
  for(size_t i = 0; i < count; i ++)
  {
    if(params[i]->values->type != vt_string &&
        params[i]->values->type != vt_object)
      die("%s:%u:Can not concat other values than object and strings",
          lang_node->filename, lang_node->line);
    len += lens[i] = strlen(params[i]->values->string);
    deps = var_dep_lists_merge(deps, params[i]->deps);
  }
  // Allocate memory for the result
  struct variable *result = var_alloc(tmp_pool, 1, temporary);
  char *res_str;
  *result->values = (struct var_value) {
    .type = vt_string,
    .string = res_str = strpool_str_get(pool, len),
    .str_temporary = temporary
  };
  result->deps = deps;
  // Copy the string parts
  size_t pos = 0;
  for(size_t i = 0; i < count; i ++)
  {
    memcpy(res_str + pos, params[i]->values->string, lens[i]);
    pos += lens[i];
  }
  return result;
}

static struct function string_concat = {
  .pcount = -1,   // Variable number of arguments
  .code = string_concat_fun,
  .params = {
    { .type = pt_single }, { .type = pt_stop }
  },
  .compiled = &var_empty
};

static struct variable *assignment_get_content(struct context *context, struct lang_node *assignment,
    struct mempool_core *tmp_pool, bool temporary)
{
  assert(assignment->block_scan);
  // FIXME This is not right, this will not work when there is an object with subblocks
  // (It will not do anything bad if the REM file is right, but will allow errors in it leading into strange
  // behaviour).
  struct context *current = assignment->block_scan->global_context ? assignment->block_scan->global_context : context;
  if(assignment->type == ln_assign_direct || assignment->type == ln_assign_alter)
  {
    assert(langlist_size(assignment) == 2);
    return expr_evaluate(current, langlist_last(assignment), tmp_pool, temporary);
  }
  else
  {
    return expr_evaluate(current, assignment, tmp_pool, temporary);
  }
}

// TODO Dependencies on files and variables, merge them
struct variable *expr_evaluate_ext(struct context *context, struct context *parent, struct lang_node *expr,
    struct mempool_core *tmp_pool, bool temporary, bool need_result, bool object_toplevel)
{
  switch(expr->type)
  {
    case ln_none:   // Never happens: can not get out
    case ln_block:  // Never happens: no-one really evaluates blocks
    case ln_import: // Never happens: Too simple to evaluate
      assert(0);
    case ln_string:
    { // Just put the string inside
      struct variable *var = var_alloc(tmp_pool, 1, temporary);
      var->values[0] = (struct var_value) {
        .type = vt_string,
        .string = expr->data
      };
      var->deps = var_dep_file(expr->filename);
      return done(var);
    }
    case ln_export:
      if(context->function)
      {
        // TODO Lists too, please ;-)
        assert(expr->children_head && !expr->children_head->next && expr->children_head->type == ln_variable);
        struct var_holder *holder = context_var_search(context, expr->children_head->data, false);
        if(holder)
          holder->exported = true;
        else
        {
          holder = context_var_search(context, expr->children_head->data, true);
          context_var_set(context, expr->children_head->data, holder ? holder->variable : &var_empty, false,
              false)->exported = true;
        }
      }
      goto FLAG;
    case ln_inherited:
      context = context_parent_first(context);
      // No break; - fall trough
    // Just commands, pass their inside
    case ln_parent:
    case ln_descend:
    // Just flags used elsewhere, pass the inside
    case ln_global:
    case ln_nocache:
    FLAG:
      // Exactly one flagged variable or one expression
      assert(expr->children_head && !expr->children_head->next);
      // Just pass it
      return expr_evaluate_ext(context, parent, expr->children_head, tmp_pool,
          temporary, need_result, object_toplevel);
    // Result of direct assignment is the right side
    case ln_assign_direct:
      if(context->function)
      {
        struct lang_node *var_name = expr->children_head;
        assert(var_name->type == ln_variable);  // TODO Other types, like subscript, array concat, etc
        struct lang_node *value_expr = var_name->next;
        assert(!value_expr->next);
        struct var_holder *holder = context_var_search(context, var_name->data, false);
        struct variable *value = expr_evaluate_ext(context, parent, value_expr, tmp_pool,
            holder && holder->ref_count > 1 ? false : temporary, true, true);
        if(holder)
          holder->variable = value;
        else
          context_var_set(context, var_name->data, value, false, false);
        return value;
      }
      else
      {
        return expr_evaluate_ext(context, parent, expr->children_head->next, tmp_pool, temporary, need_result,
            object_toplevel);
      }
    case ln_array_concat:
    { // Compute all sub-expressions and take results of all of them
      // Prepare space for the results
      size_t sub_count = langlist_size(expr);
      struct variable *subs[sub_count];
      struct subs_eval data = {
        .subs = subs,
        .context = context,
        .pool = tmp_pool
      };
      // Compute the sub-expressions
      langlist_walk(expr, &data, (langlist_walk_func) subs_eval);
      // Construct the result of them
      struct variable *result = var_alloc(tmp_pool, data.val_count, temporary);
      size_t pos = 0;
      for(size_t i = 0; i < sub_count; i ++)
      {
        memcpy(result->values + pos, subs[i]->values, subs[i]->count * sizeof *subs[i]->values);
        pos += subs[i]->count;
      }
      if(!temporary)
        var_values_make_persistent(result);
      result->deps = data.deps;
      return done(result);
    }
    case ln_string_concat:
    {
      // Compute the subs
      size_t sub_count = langlist_size(expr);
      struct variable *subs[sub_count];
      struct subs_eval data = {
        .subs = subs,
        .context = context,
        .pool = tmp_pool
      };
      langlist_walk(expr, &data, (langlist_walk_func) subs_eval);
      return function_apply(&string_concat, sub_count, subs, context, parent, tmp_pool, temporary, expr,
          object_toplevel);
    }
    case ln_variable:
      return variable_compute(context, expr->data, tmp_pool);
    case ln_variable_escape:
    {
      struct variable *original = variable_compute(context, expr->data, tmp_pool);
      struct variable *result = var_alloc(tmp_pool, 1, temporary);
      result->deps = original->deps;
      char *escaped = array_escape(original), *result_str;
      result->values[0] = (struct var_value) {
        .type = vt_string,
        .str_temporary = temporary,
        .string = result_str = mempool_alloc(pool_choose(tmp_pool, temporary), strlen(escaped) + 1)
      };
      strcpy(result_str, escaped);
      return done(result);
    }
    case ln_function:
    {
      struct variable *fun = variable_compute(context, expr->data, tmp_pool);
      if(fun->undefined)
        die("%s:%u:Function %s is not defined", expr->filename, expr->line, expr->data);
      if(!fun->count)
        return &fun->deps->var_empty;
      for(size_t i = 0; i < fun->count; i ++)
        if(fun->values[i].type != vt_function && fun->values[i].type != vt_method && fun->values[i].type != vt_class)
          die("%s:%u:Function %s is not appliable (not a function, method or class", expr->filename, expr->line,
              expr->data);
      size_t pcount = langlist_size(expr);
      struct variable *params[pcount];
      langlist_walk(expr, &(struct subs_eval) {
          .subs = params,
          .context = context,
          .pool = tmp_pool,
        }, (langlist_walk_func) subs_eval);
      struct variable *results[fun->count];
      for(size_t i = 0; i < fun->count; i ++)
        results[i] = function_apply(fun->values[i].function, pcount, params, context, parent, tmp_pool,
          temporary || (fun->count > 1), expr, object_toplevel);
      if(fun->count == 1)
        return results[0];
      else
      {
        size_t count = 0;
        for(size_t i = 0; i < count; i ++)
          count += results[i]->count;
        struct variable *result = var_alloc(tmp_pool, count, temporary);
        size_t pos = 0;
        for(size_t i = 0; i < count; i ++)
        {
          memcpy(result->values + pos, results[i]->values, results[i]->count * sizeof *result->values);
          pos += results[i]->count;
        }
        return done(result);
      }
    }
    case ln_class_def:
    {
      struct variable *class = var_alloc(tmp_pool, 1, temporary);
      class_build(expr, context, tmp_pool, temporary, class->values, class);
      class->deps = var_dep_file(expr->filename);
      return done(class);
    }
    case ln_fun_def:
    {
      struct variable *result = var_alloc(tmp_pool, 1, temporary);
      function_build(expr, context, result, tmp_pool, temporary, result->values);
      result->deps = var_dep_file(expr->filename);
      return done(result);
    }
    case ln_shell:
    {
      // It should have exactly one child, the array or string.
      // If it gets out of parser differently, something is wrong.
      assert(expr->children_head && !expr->children_head->next);
      struct variable *child = expr_evaluate(context, expr->children_head, tmp_pool, true);
      if(child->count != 1 || child->values[0].type != vt_string)
        die("%s:%u:Invalid command (must be single string value)", expr->filename, expr->line);
      const char *cmd_text = child->values[0].string;
      struct variable *echo_mode = variable_compute(context, "ECHO_MODE", tmp_pool);
      if(echo_mode->count > 1)
        diee("ECHO_MODE with multiple values");
      if(echo_mode->count && echo_mode->values[0].type != vt_string)
        diee("ECHO_MODE not a string");
      if(!echo_mode->count || strcmp(echo_mode->values[0].string, "SILENT")) // TODO more echo modes?
        puts(cmd_text);
      struct var_dep_list *deps = child->deps;
      struct variable *workdir = variable_compute(context, "WORKDIR", tmp_pool);
      assert(workdir->count == 0 || (workdir->count == 1 && workdir->values[0].type == vt_string));
      const char *result = shell_exec(cmd_text, need_result, pool_choose(tmp_pool, temporary),
          workdir->count ? workdir->values[0].string : NULL);
      if(need_result) {
        struct variable *variable = var_alloc(tmp_pool, 1, temporary);
        *variable->values = (struct var_value) {
          .type = vt_string,
          .str_temporary = temporary,
          .string = result
        };
        variable->deps = deps;
        return done(variable);
      }
      else
      {
        return &deps->var_empty;
      }
      // TODO
    }
    default:
      assert(0);  // Not implemented
  }
}

struct variable *expr_evaluate(struct context *context, struct lang_node *node, struct mempool_core *tmp_pool,
    bool temporary)
{
  return expr_evaluate_ext(context, context, node, tmp_pool, temporary, true, true);
}

static struct variable processing = {
  .state = vs_processing
};

#define STACK_PREFIX(X) var_buffer_##X
#define STACK_TYPE struct variable *
#include <common/stack.h>

struct var_process
{
  bool need_absolute;
  char nonabsolute;
  struct var_buffer_stack inserts, deletes, *current;
  struct mempool_core *tmp_pool;
  struct context *context;
  struct var_dep_list *deps;
};

static void var_process(struct node_bind *bind, struct var_process *data)
{
  if(data->need_absolute == bind->absolute && (bind->absolute || data->nonabsolute == *bind->node->data))
  {
    var_buffer_push(data->current, assignment_get_content(data->context, bind->node, data->tmp_pool, true));
    data->deps = var_dep_lists_merge(data->deps, var_dep_file(bind->node->filename));
  }
}

struct variable *variable_compute(struct context *context, const char *name, struct mempool_core *tmp_pool)
{
  struct var_holder *holder = context_var_search(context, name, true);
  if(holder)
  {
    switch(holder->variable->state)
    {
      case vs_processing:
        die("Circular variable dependency (%s)", name);
      case vs_computed:
        dprint("Providing already computed variable %s", name);
        return holder->variable;
      case vs_cache:
        dprint("Looking for variable %s in cache", name);
        if(holder->variable = cache_var_check(holder->variable, false))
          return holder->variable;
        // The value is out-of-date, compute a new one
      case vs_untouched:
      {
        dprint("Computing variable %s", name);
        struct var_process process = {
          .need_absolute = true,
          .nonabsolute = '+',
          .tmp_pool = tmp_pool,
          .context = context,
        };
        var_buffer_init(&process.inserts);
        var_buffer_init(&process.deletes);
        process.current = &process.inserts;
        struct mempool_core mempool;
        if(!tmp_pool)
        {
          mempool_core_init(&mempool, 0, true);
          process.tmp_pool = &mempool;
        }
        holder->variable = &processing;
        switch(node_bind_size(&holder->creators))
        {
          case 0:
            return holder->variable = &var_undef;
          case 1:
            var_buffer_free(&process.inserts);
            var_buffer_free(&process.deletes);
            struct variable *result = holder->variable = done(assignment_get_content(context, node_bind_first(
                &holder->creators)->node, process.tmp_pool, false));
            for(size_t i = 0; i < result->count; i ++)
              assert(!result->values[i].str_temporary && !result->values[i].fun_temporary);
            return holder->variable = result;
        }
        node_bind_walk(&holder->creators, &process, (node_bind_walk_func) var_process);
        assert(process.inserts.size <= 1);
        process.need_absolute = false;
        node_bind_walk(&holder->creators, &process, (node_bind_walk_func) var_process);
        process.nonabsolute = '-';
        process.current = &process.deletes;
        node_bind_walk(&holder->creators, &process, (node_bind_walk_func) var_process);
        // TODO remove the merged ones somehow
        size_t count = 0;
        for(size_t i = 0; i < process.inserts.size; i ++)
          count += process.inserts.data[i]->count;
        struct variable *result = var_alloc(context->context_pool, count, false);
        result->deps = process.deps;
        size_t pos = 0;
        for(size_t i = 0; i < process.inserts.size; i ++)
        {
          memcpy(result->values + pos, process.inserts.data[i]->values, process.inserts.data[i]->count *
              sizeof *result->values);
          pos += process.inserts.data[i]->count;
          result->deps = var_dep_lists_merge(result->deps, process.inserts.data[i]->deps);
        }
        // TODO remove duplicit ones
        assert(!process.deletes.size); // TODO implemenet removing of values
        if(!tmp_pool)
          mempool_core_clean(&mempool);
        var_buffer_free(&process.deletes);
        var_buffer_free(&process.inserts);
        var_values_make_persistent(result);
        return holder->variable = done(result);
      }
    }
  }
  else
  {
    dprint("Variable %s not found, treating as empty", name);
    // TODO check cache and command line
    return &var_dep_file("")->var_undef;
  }
  assert(0);  // GCC complains it can get here, but how? :-O
}
