/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "context.h"
#include "identifier.h"
#include "block.h"

#include <common/mempool_core.h>
#include <common/config.h>
#include <cache/context.h>

#include <assert.h>
#include <string.h>

#define RECYCLER_PREFIX(X) context_recycler_##X
#define RECYCLER_TYPE struct context
#define RECYCLER_NEXT cache_next
#include <common/recycler.h>

static struct context_recycler_t context_recycler;

#define RECYCLER_PREFIX(X) holder_recycler_##X
#define RECYCLER_TYPE struct var_holder
#define RECYCLER_NEXT next
#include <common/recycler.h>

static struct holder_recycler_t holder_recycler;

#define RECYCLER_PREFIX(X) plug_recycler_##X
#define RECYCLER_TYPE struct context_plug
#define RECYCLER_NEXT next
#include <common/recycler.h>

static struct plug_recycler_t plug_recycler;
struct context *context_toplevel;

static void var_provide(struct strlist *define, struct mempool_core *tmp_pool)
{
  char *name = define->string;
  char *value = index(define->string, '=');
  if(value)
    *value ++ = '\0';
  struct variable *var;
  if(!value)
  {
    context_var_set(context_toplevel, name, &var_dep_file("")->var_undef, true, false);
    return;
  }
  else if(define->common)
  {
    var = var_alloc(tmp_pool, 1, false);
    var->values->string = value;
  }
  else
  {
    char *tok, *tok_context;
    struct strlist vals = {};
    size_t count = 0;
    while(tok = strtok_r(value, ",", &tok_context))
    {
      value = NULL;
      struct strlist *new = mempool_alloc(tmp_pool, sizeof *new);
      new->string = tok;
      strlist_prepend(&vals, new);
      count ++;
    }
    var = var_alloc(tmp_pool, count, false);
    for(struct strlist *pos = vals.next; pos; pos = pos->next)
      var->values[-- count].string = pos->string;
  }
  for(size_t i = 0; i < var->count; i ++)
    var->values[i].type = vt_string;
  var->state = vs_computed;
  var->deps = var_dep_file("");
  context_var_set(context_toplevel, name, var, true, false);
}

static void var_cpy(struct var_trie_node *node, const char *name, void *unused)
{
  struct variable *var = cache_var_check(node->data->variable, strcmp(name, "TARGETS") && strcmp(name, "ACTIONS"));
  if(var)
    context_var_set(context_toplevel, name, var, true, false);
}

void lang_context_init(struct mempool_core *tmp_pool)
{
  context_recycler_init(&context_recycler, &persistent_pool);
  holder_recycler_init(&holder_recycler, &persistent_pool);
  plug_recycler_init(&plug_recycler, &persistent_pool);
  context_toplevel = context_get(&persistent_pool);
  context_toplevel->builtin = true;
}

void context_toplevel_load(struct mempool_core *tmp_pool)
{
  struct context *cached = cache_context_toplevel();
  if(cached)
    var_trie_walk(&cached->variables, NULL, var_cpy);
  // FIXME extended values too
  strlist_walk(&config_defines, tmp_pool, (strlist_walk_func) var_provide);
  static struct variable untouched = {
    .state = vs_untouched
  };
  if(!context_var_search(context_toplevel, "TARGETS", false))
    context_var_set(context_toplevel, "TARGETS", &untouched, false, false)->global = true;
  if(!context_var_search(context_toplevel, "ACTIONS", false))
    context_var_set(context_toplevel, "ACTIONS", &untouched, false, false)->global = true;
}

struct context *context_get(struct mempool_core *pool)
{
  struct context *result = context_recycler_get(&context_recycler);
  *result = (struct context) { };
  var_trie_init(&result->variables, pool);
  result->context_pool = pool;
  return result;
}

void context_return(struct context *context)
{
  context_recycler_put(&context_recycler, context);
}

struct parent_search
{
  struct var_holder *result;
  const char *name;
};

static void parent_search(struct context_plug *plug, struct parent_search *data)
{
  if(!data->result)
    data->result = context_var_search(plug->context, data->name, true);
}

struct var_holder *context_var_search(struct context *context, const char *name, bool parents)
{
  // Look for local variable
  struct var_holder *result = var_trie_find(&context->variables, name);
  if(!result && parents)
  { // Search trough parents
    struct parent_search data = { .name = name };
    plug_list_walk(&context->parents, &data, (plug_list_walk_func) parent_search);
    result = data.result;
  }
  return result;
}

void context_var_reference(struct context *context, const char *name, struct var_holder *holder_new, bool overwrite)
{
  struct var_holder *holder = context_var_search(context, name, false);
  if(holder == holder_new)
    return;
  assert((holder_new->variable->state != vs_computed && holder_new->variable->state != vs_cache) || !holder ||
      (holder->variable->state != vs_computed && holder->variable->state != vs_cache ||
       holder->variable == holder_new->variable));
  assert(overwrite || !holder);
  if(holder && !-- holder->ref_count)
    holder_recycler_put(&holder_recycler, holder);
  holder_new->ref_count ++;
  var_trie_insert(&context->variables, name, holder_new);
}

struct var_holder *context_var_set(struct context *context, const char *name, struct variable *value, bool overwrite,
    bool subscripted)
{
  struct var_holder *holder = holder_recycler_get(&holder_recycler);
  *holder = (struct var_holder) {
    .variable = value,
    .ref_count = 1,
    .subscripted = subscripted
  };
  var_trie_insert(&context->variables, name, holder);
  return holder;
}

static struct variable var_untouched = {
  .state = vs_untouched
};

static void context_prepare_out(struct identifier *identifier, struct context *context)
{
  struct var_holder *holder = context_var_search(context, identifier->name, false);
  if(!holder)
    holder = context_var_set(context, identifier->name, &var_untouched, false, false);
  holder->global = identifier->global;
  holder->exported = identifier->exported;
  holder->creators = identifier->creators; // TODO Is this OK every time?
                                           // Is this OK to ommit when from cache?
}

void context_prepare(struct context *context)
{
  if(context->function)
  {
  }
  else
    identifier_list_walk(&context->block->out, context, (identifier_list_walk_func) context_prepare_out);
}

static void globals_identify(struct var_trie_node *node, const char *name, struct context *child)
{
  struct var_holder *variable = node->data;
  if(variable->global && !child->builtin)
  {
    // TODO check for collisions in assignment
    // TODO merge sources
    struct var_holder *hol = context_var_search(child, name, false);
    if(hol == variable)
      return;
    if(hol)
      node_bind_concat_append(&variable->creators, &hol->creators);
    context_var_reference(child, name, variable, true);
  }
}

void context_var_set_global(struct context *context, const char *name)
{
  struct var_holder *holder = context_var_search(context, name, false);
  if(!holder)
    holder = context_var_set(context, name, &var_untouched, false, false);
  holder->global = true;
}

struct exports_identify
{
  struct context *child, *parent;
};

static void exports_identify(struct var_trie_node *node, const char *name, struct exports_identify *contexts)
{
  struct var_holder *variable = node->data;
  if(variable->exported)
  {
    struct var_holder *par_hold = context_var_search(contexts->parent, name, false);
    if(par_hold)
    {
      if(par_hold == variable) // Already merged, maybe from cache
      {
        variable->exported = false;
        return;
      }
      assert(variable->ref_count == 1);
      par_hold->global = par_hold->global || variable->global;
      node_bind_concat_append(&par_hold->creators, &variable->creators);
      assert(par_hold->variable->state == vs_untouched || par_hold->variable->state == vs_cache);
      par_hold->variable = variable->variable;
      if(!contexts->child->builtin)
        context_var_reference(contexts->child, name, par_hold, true);
    }
    else
    {
      if(!contexts->child->builtin)
        variable->exported = false;
      context_var_reference(contexts->parent, name, variable, true);
    }
  }
}

void contexts_plug(struct context *parent, struct context *child)
{
  if(parent == child)
    return;
  // Store the parent
  struct context_plug *plug = plug_recycler_get(&plug_recycler);
  plug->context = parent;
  plug_list_prepend(&child->parents, plug);
  // TODO Some special handling with function context?
  // Export variables from child to the parent
  var_trie_walk(&child->variables, &(struct exports_identify) { .parent = parent, .child = child },
      (var_trie_walk_func) exports_identify);
  // Identify global variables with the ones in the child
  var_trie_walk(&parent->variables, child, (var_trie_walk_func) globals_identify);
}

struct context *context_parent_first(struct context *context)
{
  if(context->parents.next)
    return context->parents.next->context;
  else
    return NULL;
}

void context_raw_block(struct block_scan_result *result, struct lang_node *node)
{
  // Look into the cache if we have one
  if(!(result->global_context = cache_context_get(node->filename, node->token)))
    result->global_context = context_get(&persistent_pool);
  result->global_context->block = result;
  context_prepare(result->global_context);
  result->global_context->dep_list = var_dep_file(node->filename);
  assert(!result->global_context->object);
  // Mark for cache storage
  cache_context_put(result->global_context, NULL);
}

static void export_func(struct var_trie_node *node, const char *name, struct context *parent)
{
  if(node->data->exported && node->data->ref_count == 1)
  {
    struct var_holder *var_par = context_var_search(parent, name, false);
    if(var_par)
      var_par->variable = var_make_persistent(node->data->variable);
    else
      context_var_set(parent, name, var_make_persistent(node->data->variable), false, false);
  }
}

void context_export_func(struct context *context)
{
  var_trie_walk(&context->variables, context_parent_first(context), (var_trie_walk_func) export_func);
}

void lang_context_restart(void)
{
  context_recycler_clean(&context_recycler);
  holder_recycler_clean(&holder_recycler);
  plug_recycler_clean(&plug_recycler);
  context_toplevel = NULL;
}
