/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "function.h"
#include "node.h"
#include "var.h"
#include "context.h"
#include "compute.h"

#include <common/debug.h>
#include <common/mempool_core.h>
#include <string.h>

struct dup_def
{
  size_t amount;
  size_t pos;
  size_t index;
};

#define STACK_PREFIX(X) dup_def_##X
#define STACK_TYPE struct dup_def
#include <common/stack.h>

struct variable *function_apply(struct function *function, size_t param_count, struct variable **params,
    struct context *context, struct context *parent, struct mempool_core *tmp_pool, bool temporary,
    struct lang_node *lang_node, bool object_toplevel)
{
  if(function->pcount != -1 && (size_t) function->pcount != param_count)
    die("%s:%u:Invalid number of parameters%s%s, expects %d, got %zu", lang_node->filename, lang_node->line,
        lang_node->data ? " for function " : "", lang_node->data ? lang_node->data : "", function->pcount, param_count);
  // Count how many times we duplicit the function
  size_t dup_count = 1;
  struct param_def *pdef = function->params;
  struct dup_def_stack dup_def_stack;
  dup_def_init(&dup_def_stack);
  for(size_t i = 0; i < param_count; i ++)
  {
    if(pdef->type == pt_stop)
      pdef --;
    if(pdef->type == pt_single && params[i]->count != 1)
    {
      dup_count *= params[i]->count;
      dup_def_push(&dup_def_stack, (struct dup_def) {
          .pos = i,
          .amount = params[i]->count
        });
    }
    pdef ++;
  }
  struct variable *result;
  switch(dup_count)
  {
    case 0: // No functions called, it is empty
      result = function->compiled->deps ? &function->compiled->deps->var_empty : &var_empty;
      break;
    case 1:
      result = function->code(params, param_count, context, parent, tmp_pool, temporary, lang_node, function,
          object_toplevel);
      result->deps = var_dep_lists_merge(result->deps, function->compiled->deps);
      break;
    default: // OK, we multiply
    {
      // Get a copy of the parameters
      struct variable *params_mod[param_count];
      memcpy(params_mod, params, param_count *sizeof *params);
      char par_buf[dup_def_stack.size * (sizeof **params + sizeof *(*params)->values)];
      // Prepare single-value variables to the right places
      for(size_t i = 0; i < dup_def_stack.size; i ++)
      {
        params_mod[dup_def_stack.data[i].pos] = (struct variable *) (par_buf + i * (sizeof **params +
              sizeof *(*params)->values));
        *params_mod[dup_def_stack.data[i].pos] = (struct variable) {
          .state = vs_computed,
          .count = 1,
          .temporary = true,
          .deps = params[i]->deps
        };
      }
      // Compute the separate results
      struct variable *results[dup_count];
      size_t result_pos = 0;
      size_t depth = 0;
      struct var_dep_list *deps = function->compiled->deps;
      while(depth || dup_def_stack.data->index < dup_def_stack.data->amount)
      {
        // All is specified, run the function once
        if(depth == dup_def_stack.size)
        {
          results[result_pos] = function->code(params_mod, param_count, context, parent, tmp_pool, true, lang_node,
              function, object_toplevel);
          depth --; // Lower one level to change it
          deps = var_dep_lists_merge(deps, results[result_pos ++]->deps);
        }
        // Not at the end of them at this level
        else if(dup_def_stack.data[depth].index < dup_def_stack.data[depth].amount)
        {
          size_t pos = dup_def_stack.data[depth].pos;
          // Copy the right parameter, increase the index for the next time
          // and move to the higher level to check
          *params_mod[pos]->values = params[pos]->values[dup_def_stack.data[depth ++].index ++];
        }
        // This level is at the end -> go lower to change it and reset this
        else
          dup_def_stack.data[depth --].index = 0;
      }
      // Merge the results
      size_t result_count = 0;
      // Compute the size of result
      for(size_t i = 0; i < dup_count; i ++)
        result_count += results[i]->count;
      // Allocate the result
      result = var_alloc(tmp_pool, result_count, temporary);
      // Copy the values
      size_t value_pos = 0;
      for(size_t i = 0; i < dup_count; i ++)
      {
        memcpy(result->values + value_pos, results[i]->values, results[i]->count * sizeof *results[i]->values);
        value_pos += results[i]->count;
      }
      if(!temporary)
        var_values_make_persistent(result);
      result->deps = deps;
    }
  }
  dup_def_free(&dup_def_stack);
  result->state = vs_computed;
  result->deps = var_dep_lists_merge(result->deps, var_dep_file(lang_node->filename));
  return result;
}

struct line_eval
{
  struct mempool_core *tmp_pool;
  struct context *context;
};

void line_eval(struct lang_node *command, struct line_eval *data)
{
  expr_evaluate_ext(data->context, data->context, command, data->tmp_pool, true, false, true);
}

struct variable *function_user_eval(struct variable **params, size_t count, struct context *parent_context,
    struct context *parent, struct mempool_core *tmp_pool, bool temporary, struct lang_node *lang_node,
    struct function *function, bool object_toplevel)
{
  assert(count == (size_t) function->pcount);
  // Get context for this function
  struct context *context = context_get(tmp_pool);
  for(size_t i = 0; i < count; i ++)
    context_var_set(context, function->params[i].name, params[i], false, false);
  context->dep_list = function->compiled->deps;
  context->function = true;
  context_prepare(context);
  contexts_plug(function->func_context, context);
  assert(function->block->type == ln_block);
  langlist_walk(function->block, &(struct line_eval) {
      .tmp_pool = tmp_pool,
      .context = context
    }, (langlist_walk_func) line_eval);
  struct var_holder *result_holder = context_var_search(context, "RESULT", false);
  struct variable *result;
  context_export_func(context);
  if(result_holder)
  {
    result = result_holder->variable;
    result->deps = var_dep_lists_merge(function->compiled->deps, result->deps);
    if(!temporary)
      result = var_make_persistent(result);
  }
  else
    result = &function->compiled->deps->var_empty;
  // The context is not needed, it is for the function only
  context_return(context);
  return result;
}

static void fun_fill(struct lang_node *node, struct param_def **param)
{
  switch(node->type)
  {
    case ln_param_multi:
    case ln_param_single:
      *(*param) ++ = (struct param_def) {
        .name = node->data,
        .type = node->type == ln_param_multi ? pt_multi : pt_single
      };
      break;
    case ln_block:
      break;
    default:
      assert(0);
  }
}

void function_build(struct lang_node *expr, struct context *context, struct variable *compiled,
    struct mempool_core *tmp_pool, bool temporary, struct var_value *value)
{
  // Params. The last one is the block.
  size_t param_count = langlist_size(expr) - 1;
  struct function *function = mempool_alloc(pool_choose(tmp_pool, temporary), sizeof *function + param_count *
      sizeof *function->params);
  *function = (struct function) {
    .pcount = param_count,
    .code = function_user_eval,
    .block = langlist_last(expr),
    .func = expr,
    .compiled = compiled,
    .func_context = context
  };
  *value = (struct var_value) {
    .type = context->object ? vt_method : vt_function,
    .fun_temporary = temporary,
    .object = context->object,
    .function = function
  };
  struct param_def *param_def = function->params;
  langlist_walk(expr, &param_def, (langlist_walk_func) fun_fill);
}
