/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "var.h"

#include <common/mempool_core.h>
#include <common/stringpool.h>

#include <string.h>

struct variable *var_alloc(struct mempool_core *tmp_pool, size_t count, bool temporary)
{
  struct variable *result = mempool_alloc(pool_choose(tmp_pool, temporary),
      sizeof *result + count * sizeof *result->values);
  *result = (struct variable) {
    .count = count,
    .state = vs_processing,
    .temporary = temporary
  };
  return result;
}

struct variable var_empty = {
  .state = vs_computed
};

struct variable var_undef = {
  .state = vs_computed,
  .undefined = true
};

void var_values_make_persistent(struct variable *variable)
{
  for(size_t i = 0; i < variable->count; i ++)
  {
    if(variable->values[i].function && variable->values[i].fun_temporary)
    {
      size_t size = sizeof *variable->values[i].function + (variable->values[i].function ?
          variable->values[i].function->pcount : 0)* sizeof *variable->values[i].function->params;
      struct function *func = mempool_alloc(&persistent_pool, size);
      memcpy(variable->values[i].function, func, size);
    }
    if(variable->values[i].str_temporary)
      variable->values[i].string = strpool_dup(&strpool, variable->values[i].string);
    variable->values[i].fun_temporary = variable->values[i].str_temporary = false;
  }
}

struct variable *var_make_persistent(struct variable *variable)
{
  if(variable->temporary)
  {
    struct variable *result = var_alloc(NULL, variable->count, false);
    memcpy(result, variable, sizeof *variable + variable->count * sizeof *variable->values);
    result->temporary = false;
    var_values_make_persistent(result);
    return result;
  }
  else
  {
    var_values_make_persistent(variable);
    return variable;
  }
}

#include "var_dep.h"

#define RECYCLER_PREFIX(X) dep_recycler_##X
#define RECYCLER_TYPE struct var_dep
#define RECYCLER_NEXT next
#include <common/recycler.h>

struct var_dep_chain var_dep_single, var_dep_merged;
static struct dep_recycler_t dep_recycler;

void lang_var_init(void)
{
  dep_recycler_init(&dep_recycler, &persistent_pool);
}

struct var_dep_file_scan
{
  const char *filename;
  struct var_dep_list *result;
};

static void var_dep_file_scan(struct var_dep_list *list, struct var_dep_file_scan *data)
{
  if(!strcmp(list->head->filename, data->filename))
    data->result = list;
}

struct var_dep_list *var_dep_file(const char *filename)
{
  // Do I have this one already?
  struct var_dep_file_scan data = {
    .filename = filename
  };
  dep_chain_walk(&var_dep_single, &data, (dep_chain_walk_func) var_dep_file_scan);
  if(data.result)
    return data.result;
  // Not found, construct a new one and store it for next time
  struct var_dep_list *result = mempool_alloc(&persistent_pool, sizeof *result);
  struct var_dep *dep = dep_recycler_get(&dep_recycler);
  dep_list_append(result, dep);
  *dep = (struct var_dep) { .filename = filename };
  dep_chain_append(&var_dep_single, result);
  result->var_empty = var_empty;
  result->var_empty.deps = result;
  result->var_undef = result->var_empty;
  result->var_undef.undefined = true;
  return result;
}

static void filter_build(struct var_dep_list *list, struct var_dep_chain *result)
{
  dep_filter_append(result, list);
  list->next_test = list->head;
}

struct filter_run
{
  const char *filename;
  struct var_dep_chain filter;
};

static void filter_run(struct var_dep_list *list, struct filter_run *data)
{
  // Leave only the lists with the propper filename here
  if(list->next_test && !strcmp(list->next_test->filename, data->filename))
  {
    list->next_test = list->next_test->next;
    dep_filter_append(&data->filter, list);
  }
}

static void filter_finish(struct var_dep_list *list, struct filter_run *data)
{
  if(!list->next_test)
    dep_filter_append(&data->filter, list);
}

static void dep_clean(struct var_dep *dep, void *unused)
{
  dep_recycler_put(&dep_recycler, dep);
}

struct var_dep_list *var_dep_lists_merge(struct var_dep_list *_1, struct var_dep_list *_2)
{
  if(!_1)
    return _2;
  if(!_2)
    return _1;
  assert(_1->head && _2->head); // No empty dep lists should exist
  // Merge the same lists - not interesting
  if(_1 == _2)
    return _1;
  struct var_dep_list new = {};
  struct var_dep *d1 = _1->head, *d2 = _2->head;
  // Prepare list of still valid chains
  struct var_dep_chain filter = {};
  dep_chain_walk(&var_dep_merged, &filter, (dep_chain_walk_func) filter_build);
  // Merge, filter trough the chains as you go
  while(d1 || d2)
  {
    // Make sure d1 < d2, so I do not have to care in the rest which one
    int cmp = d1 ? d2 ? strcmp(d1->filename, d2->filename) : -1 : 1;
    if(cmp > 0)
    {
      struct var_dep *tmp = d1;
      d1 = d2;
      d2 = tmp;
    }
    else if(cmp == 0)
      d2 = d2->next;
    assert(!d2 || strcmp(d1->filename, d2->filename) < 0);
    // Remove the wrong ones from the valid ones
    struct filter_run data = { .filename = d1->filename };
    dep_filter_walk(&filter, &data, (dep_filter_walk_func) filter_run);
    filter = data.filter;
    // Add the new node to the new chain
    struct var_dep *dep = dep_recycler_get(&dep_recycler);
    *dep = (struct var_dep) { .filename = d1->filename };
    dep_list_append(&new, dep);
    d1 = d1->next;
  }
  // Look if there is a matching chain
  struct filter_run result = {};
  dep_filter_walk(&filter, &result, (dep_filter_walk_func) filter_finish);
  if(result.filter.head)
  {
    // The chain exists, return this one, as it is not needed
    dep_list_walk(&new, NULL, dep_clean);
    result.filter.head->changed_resolved |= _1->changed_resolved && _2->changed_resolved;
    result.filter.head->changed = (_1->changed_resolved && _2->changed_resolved) ? (_1->changed || _2->changed) :
        result.filter.head->changed;
    return result.filter.head;
  }
  else
  {
    // Ok, not exists, finish creating it
    struct var_dep_list *list = mempool_alloc(&persistent_pool, sizeof *list);
    new.changed = _1->changed || _2->changed;
    new.changed_resolved = _1->changed_resolved && _2->changed_resolved;
    *list = new;
    dep_chain_append(&var_dep_merged, list);
    list->var_empty = var_empty;
    list->var_empty.deps = list;
    list->var_undef = list->var_empty;
    list->var_undef.undefined = true;
    return list;
  }
}

void lang_var_restart(void)
{
  var_dep_single = var_dep_merged = (struct var_dep_chain) {};
  dep_recycler_clean(&dep_recycler);
}

bool var_eq(struct variable *_1, struct variable *_2)
{
  if(_1->count != _2->count)
    return false;
  for(size_t i = 0; i < _1->count; i ++)
  {
    if(_1->values[i].type != _2->values[i].type)
      return false;
    switch(_1->values[i].type)
    {
      case vt_method:
      case vt_function:
      case vt_class:
        if(_1->values[i].function != _2->values[i].function)
          return false;
        if(_1->values[i].type != vt_method)
          break;
      case vt_object:
        if(_1->values[i].object != _2->values[i].object)
          return false;
      case vt_string:
        if(strcmp(_1->values[i].string, _2->values[i].string))
          return false;
    }
  }
  return true;
}
