/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "token.h"

#include <common/memory.h>
#include <common/stringpool.h>
#include <common/str.h>

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdbool.h>

struct token_state
{
  const char *buffer; // Input data
  char *token; // Current token
  size_t tok_len, tok_free; // Sizes of token
  unsigned line_num; // Current line number
  unsigned errors; // Number of errors encountered
  void *data; // User's data
  token_error_handler_t error_handler;
  token_handler_t handler;
  // State of the automata
  bool in_string:1;
  bool in_comment:1;
  bool in_ident:1;
  bool in_perl:1;
  bool in_oper:1;
  bool escape:1;
  bool newline:1;
  bool subst_string:1;
  bool ident_escape:1;
};

#define IS_ID_CHAR(c) ((c == '_') || ((c >= 'a') && (c <= 'z')) \
        || ((c >= 'A') && (c <= 'Z')) \
        || ((c >= '0') && (c <= '9')))

static bool is_oper_char(char c)
{
  return index(":=+-~!.<>{}()[],;\\,\"'\n#$@", c);
}

static bool is_whitespace(char c)
{
  return index(" \t\n", c);
}

static void token_warn(struct token_state *state, const char *format,
    ...) __attribute__((format(printf,2,3)));
static void token_warn(struct token_state *state, const char *format,
    ...)
{
  VAPRINTF(stage1, format);
  char *stage2 = aprintf("%d: %s\n", state->line_num, stage1);
  xfree(stage1);
  if(state->error_handler)
    state->error_handler(state->data, stage2);
  xfree(stage2);
  state->errors ++;
}

// Add a char to the end of token
static void token_push(struct token_state *state, char c)
{
  if(!state->tok_free)
    state->token = (char *) xrealloc(state->token - state->tok_len,
        2 * (state->tok_free = state->tok_len)) + state->tok_len;
  *state->token ++ = c;
  state->tok_len ++;
  state->tok_free --;
}

// Take the last from token.
static char token_pop(struct token_state *state)
{
  assert(state->tok_len);
  state->tok_len --;
  state->tok_free ++;
  return *state->token --;
}

static void token_clean(struct token_state *state)
{
  state->token -= state->tok_len;
  state->tok_free += state->tok_len;
  state->tok_len = 0;
}

static void token_pack(struct token_state *state, enum token_type type)
{
  if(type == tt_string_data && !state->tok_len)
    return;
  token_push(state, '\0');
  if(state->handler)
    state->handler(state->data, state->token - state->tok_len,
        state->tok_len - 1, type, state->line_num);
  token_clean(state);
}

static struct oper_type
{
  const char *oper;
  enum token_type type;
} opers[] = {
  { ":=", tt_operator },
  { "+=", tt_operator },
  { "-=", tt_operator },
  { ":~", tt_operator },
  { "+~", tt_operator },
  { "-~", tt_operator },
  { "!", tt_operator },
  { ".", tt_operator },
  { ":", tt_operator },
  { "<<<", tt_internal },
  { "(", tt_open_paren },
  { ")", tt_close_paren },
  { "{", tt_open_block },
  { "}", tt_close_block },
  { "[", tt_start_subscript },
  { "]", tt_end_subscript },
  { ",", tt_comma },
  { "'", tt_internal },
  { "\"", tt_internal },
  { "\\", tt_internal },
  { "\n", tt_internal },
  { ";", tt_command_end },
  { "#", tt_internal },
  { "$", tt_vartype },
  { "@", tt_vartype },
  { NULL, tt_internal }
};

static struct oper_type *oper_find(struct token_state *state,
    int end_offset)
{
  size_t len = state->tok_len + end_offset;
  for(struct oper_type *t = opers; t->oper; t ++)
    if(strlen(t->oper) == len && !memcmp(t->oper, state->token -
          state->tok_len, len))
      return t;
  return NULL;
}

static void oper_handle(struct token_state *state)
{
  struct oper_type *type = oper_find(state, 0);
  if(type)
  {
    if(type->type == tt_internal)
    {
#define OP_EQ(what) (!strcmp((what), type->oper))
      if(OP_EQ("<<<"))
        state->in_perl = true;
      else if(OP_EQ("'"))
      {
        state->in_string = true;
        token_pack(state, tt_string_start);
      }
      else if(OP_EQ("\""))
      {
        state->in_string = state->subst_string = true;
        token_pack(state, tt_string_start);
      }
      else if(OP_EQ("\\"))
        state->escape = true;
      else if(OP_EQ("\n"))
        if(state->escape)
          state->escape = false;
        else
          token_pack(state, tt_command_end);
      else if(OP_EQ("#"))
        state->in_comment = true;
      else
      {
        assert(0);
      }
#undef OP_EQ
      token_clean(state);
    }
    else
      token_pack(state, type->type);
  }
  else
  {
    token_push(state, '\0');
    token_warn(state, "Invalid operator '%s'", state->token -
        state->tok_len);
    token_clean(state);
  }
  state->in_oper = false;
}

static void char_push(struct token_state *state, char c)
{
  if(state->in_comment)
  {
    if(c == '\n')
    {
      /*
       * End skipping the comment and terminate the command
       */
      state->in_comment = false;
      char_push(state, '\n');
    }
  }
  else if(state->in_perl)
  {
    // End of perl mode
    if((c == '>') && (state->tok_len >= 2) && (!memcmp(state->token - 2,
            ">>", 2)))
    {
      state->token -= 2;
      state->tok_len -= 2;
      token_pack(state, tt_perl);
      state->in_perl = false;
    }
    else
      token_push(state, c);
  }
  else if(state->in_ident)
  {
    if(IS_ID_CHAR(c))
      token_push(state, c);
    else
    {
      state->in_ident = false;
      if(state->subst_string)
        // If it is a substitution variable, continue the string
        token_pack(state, state->ident_escape ? tt_string_var_escaped :
            tt_string_var);
      else
        token_pack(state, tt_ident);
      state->ident_escape = false;
      // Usual processing of the char
      char_push(state, c);
    }
  }
  else if(state->in_string)
  {
    if(state->escape)
    { // Unescape
      state->escape = false;
      switch(c)
      {
        case 'n': token_push(state, '\n'); break;
        case 'r': token_push(state, '\r'); break;
        case 't': token_push(state, '\t'); break;
        case 'e': token_push(state, '\e'); break;
        case '\\':
        case '\'':
        case '"':
        case '$':
        case '@':
          token_push(state, c); break;
        default:
        {
          token_warn(state, "Invalid escape character '%c'", c);
        }
      }
    }
    else if((c == '\\') && (state->subst_string))
      // Escape. Only in substitued strings.
      state->escape = true;
    else if(((c == '$') || (c == '@')) && (state->subst_string))
    { // Variable
      token_pack(state, tt_string_data);
      state->in_ident = true;
      state->ident_escape = c == '@';
    }
    else if((c == '\'' && !state->subst_string) || (c == '"' && state->subst_string))
    { // String end
      // Send out both string data and string end
      token_pack(state, tt_string_data);
      token_push(state, c);
      token_pack(state, tt_string_end);
      state->in_string = state->subst_string = false;
    }
    else
      token_push(state, c);
  }
  else if(state->in_oper)
  {
    if(is_oper_char(c))
    {
      token_push(state, c);
      // Two operators together?
      if(oper_find(state, -1) && !oper_find(state, 0))
      {
        (void) token_pop(state);
        oper_handle(state);
        char_push(state, c);
      }
    }
    else
    {
      oper_handle(state);
      char_push(state, c);
    }
  }
  else
  {
    // Start some token?
    if(state->escape && !is_whitespace(c))
    {
      state->escape = false;
      token_warn(state, "Mistplaced '\\'");
    }
    if(is_oper_char(c))
    {
      state->in_oper = true;
      token_push(state, c);
    }
    else if(IS_ID_CHAR(c))
    {
      state->in_ident = true;
      token_push(state, c);
    }
    // No token started. Only whitespace is ignored.
    else if(!is_whitespace(c))
      token_warn(state, "Invalid character '%c'", c);
  }
}

/*
 * This one should bypass differences between all ASCII based newline characters.
 * It simply translates all valid newline sequences ("\n", "\r", "\r\n" to "\n").
 */
static char newline_normalize(struct token_state *state, char c)
{
  if(c == '\n')
    if(state->newline)
    {
      // Running on DOS-like system, ignore second part of newline
      state->newline = false;
      return '\0';
    }
    else
      // Unix, pass this trough
      return '\n';
  else if(c == '\r')
  {
    /*
     * This is either DOS-like, first part of the newline
     * or Mac. Output as newline and mark it in case it is DOS.
     */
    state->newline = true;
    return '\n';
  }
  else
  {
    state->newline = false;
    return c;
  }
}

/*
 * Process one character of input.
 */
static void char_process(struct token_state *state)
{
  char c;
  if((c = newline_normalize(state, *state->buffer ++)))
    char_push(state, c);
  if(c == '\n')
    state->line_num ++;
}

struct token_state *token_create(void *data, token_error_handler_t
    error_handler, token_handler_t handler)
{
  struct token_state *state = xmalloc(sizeof *state);
  *state = (struct token_state) {
    .token = xmalloc(100),
    .tok_free = 100,
    .data = data,
    .error_handler = error_handler,
    .handler = handler,
    .line_num = 1
  };
  return state;
}

void token_process(struct token_state *state, const char *buffer,
    size_t len)
{
  state->buffer = buffer;
  const char *end = buffer + len;
  while(state->buffer < end)
    char_process(state);
}

unsigned token_destroy(struct token_state *state)
{
  if(state->escape || state->in_string || state->in_perl)
    token_warn(state, "Unterminated input");
  else
  {
    // Must be two of them, to push the first one trough
    char_push(state, '\n');
    char_push(state, '\n');
  }
  unsigned result = state->errors;
  xfree(state->token - state->tok_len);
  xfree(state);
  return result;
}

#ifdef TEST

#include <common/msg.h>

void error_handler(void *data, const char *msg)
{
  (void) data;
  fwrite(msg, 1, strlen(msg), stderr);
}

void handler(void *data, const char *token, size_t len,
    enum token_type type)
{
  (void) data;
  (void) len;
  printf("token: %i: '%s'\n", (int) type, token);
}

int main(int argc, const char *argv[])
{
  gdb_catch_signals();
  (void) argc;
  (void) argv;
  struct token_state *state = token_create(NULL, &error_handler,
      &handler);
  size_t amount;
  char buffer[1024];
  while((amount = fread(buffer, 1, 1024, stdin)))
    tokenize_data(state, buffer, amount);
  (void) token_destroy(state);
  return 0;
}

#endif
