/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

//TODO Many asserts here will have to be turned into syntax error checks, add more syntax checks
#include "tree.h"

#include <lang/node.h>
#include <common/memory.h>
#include <common/debug.h>
#include <common/stringpool.h>
#include <common/str.h>
#include <common/config.h>

#define STACK_PREFIX(X) lang_##X
#define STACK_TYPE struct lang_node *
#include <common/stack.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

struct treeb_state
{
  struct lang_stack oper_stack, value_stack, block_stack;
  treeb_error_handler_t error_handler;
  unsigned last_token, last_value, last_open_paren;
  const char *filename;
  unsigned line;
  void *data;
  bool has_vartype:1;
  bool eat_block:1;
  bool expect_def:1;
  bool expect_params:1;
  bool is_map:1;
  bool is_extract:1;
  bool is_fun:1;
  bool is_multi:1;
  bool in_fun:1;
  bool in_def:1;
};

struct treeb_state *treeb_create(void *data,
    treeb_error_handler_t error_handler, const char *filename)
{
  struct treeb_state *result = xmalloc(sizeof *result);
  *result = (struct treeb_state) {
    .error_handler = error_handler,
    .data = data,
    .last_token = 1,
    .filename = strpool_dup(&strpool, filename)
  };
  lang_init(&result->oper_stack);
  lang_init(&result->value_stack);
  lang_init(&result->block_stack);
  lang_push(&result->oper_stack, NULL);
  treeb_insert(result, "{", 1, tt_open_block, 0);
  return result;
}

struct lang_node *treeb_destroy(struct treeb_state *state)
{
  // TODO check there is exactly one item in the blocks stack
  dprint("Destroying tree builder, sizes (value, oper, block):"
      "%u %u %u", lang_size(&state->value_stack),
      lang_size(&state->oper_stack), lang_size(&state->block_stack));
  struct lang_node *node = lang_top(&state->block_stack);
  dprinte("Parsed tree:");
#ifdef TRACE
  if(config_debug)
    lang_tree_print(node);
#endif
  lang_free(&state->block_stack);
  lang_free(&state->oper_stack);
  lang_free(&state->value_stack);
  xfree(state);
  return node;
}

static struct lang_node *node_alloc(struct treeb_state *state)
{
  struct lang_node *node = lang_node_alloc();
  node->line = state->line;
  node->filename = state->filename;
  node->token = state->last_token;
  return node;
}

static void fatal(struct treeb_state *state, const char *msg, ...)
    __attribute__((noreturn)) __attribute__((format(printf,2,3)));
static void fatal(struct treeb_state *state, const char *msg, ...)
{
  VAPRINTF(stage1, msg);
  char *stage2 = aprintf("%u: %s\n", state->line, stage1);
  xfree(stage1);
  if(state->error_handler)
    state->error_handler(state->data, stage2, 1);
#ifdef DEBUG
  abort();
#else
  exit(1);
#endif
}

// Marks into the the operator stack
static struct lang_node comma_stop, parenthesis_stop, string_stop,
    subscript_stop, space_oper;

static struct
{
  const char *name;
  unsigned short priority;
  enum lang_node_type type;
  short arity;
} opers[] = {
  { ".", 3, ln_string_concat, 2 },
  { ":", 2, ln_method, 2 },
  { "!", 1, ln_shell, 1 },
  { ":=", 1, ln_assign_direct, -2 },
  { "+=", 1, ln_assign_alter, -2 },
  { "-=", 1, ln_assign_alter, -2 },
  { ":~", 1, ln_inherit_direct, -2 },
  { "+~", 1, ln_inherit_alter, -2 },
  { "-~", 1, ln_inherit_alter, -2 },
  { "eval", 1, ln_eval, 1 },
  { "parent", 1, ln_parent, 1 },
  { "descend", 1, ln_descend, 1 },
  { "import", 1, ln_import, 1 },
  { "inherited", 4, ln_inherited, 1 },
  { "export", 4, ln_export, 1 },
  { "global", 4, ln_global, 1 },
  { "nocache", 4, ln_nocache, 1 },
  { " ", 1, ln_none, 0 },
  { NULL, 0, ln_none, 1 }
};

static unsigned short oper_consult(struct lang_node *node,
    const char *token)
{
  if(!token &&(!node || !node->data))
    return 0;// This is a stopper
  unsigned i = 0;
  if(!token && node->data)
  {
    token = node->data;
    i = node->oper;
  }
  for(; opers[i].name; i ++)
    if(!strcmp(token, opers[i].name))
    {
      if(node && !node->data)
      {
        node->data = opers[i].name;
        node->type = opers[i].type;
        node->oper = i;
      }
      return opers[i].priority;
    }
  return 0;
}

static short oper_arity(struct lang_node *node)
{
  return opers[node->oper].arity;
}

static void array_load(struct treeb_state *state,
    struct lang_node *where)
{
  struct lang_node *node;
  while((node = lang_pop(&state->value_stack)))
    langlist_prepend(where, node);
}

static void value_insert(struct treeb_state *state,
    struct lang_node *node, int syntetic);

static void operator_handle(struct treeb_state *state,
    struct lang_node *node)
{
  (void) state;
  short arity = oper_arity(node);
  dprint("Handling operator '%s' of arity %i", node->data, (int) arity);
  if(!arity)
    return;
  if(arity < 0)
  {
    arity *= -1;
    struct lang_node *newnode = node_alloc(state);
    newnode->type = ln_array_concat;
    array_load(state, newnode);
    lang_push(&state->value_stack, newnode);
  }
  while(arity --)
    langlist_prepend(node, lang_pop(&state->value_stack));
  value_insert(state, node, 1);
}

static void oper_run_until(struct treeb_state *state,
    struct lang_node *stop)
{
  // TODO Check if we are still inside the stack
  struct lang_node *node;
  while((node = lang_pop(&state->oper_stack)) != stop)
    operator_handle(state, node);
}

static void oper_run_all(struct treeb_state *state)
{
  oper_run_until(state, NULL);
  lang_push(&state->oper_stack, NULL);
}

static void oper_insert(struct treeb_state *state,
    struct lang_node *node, const char *token)
{
  unsigned short priority = node->priority = oper_consult(node, token);
  while(lang_size(&state->oper_stack) && oper_consult(lang_top(
      &state->oper_stack), NULL) > priority)
    operator_handle(state, lang_pop(&state->oper_stack));
  lang_push(&state->oper_stack, node);
  if(oper_arity(node) < 0) // Takes an array as the second argument
    lang_push(&state->value_stack, NULL);
}

static void value_insert(struct treeb_state *state,
    struct lang_node *node, int syntetic)
{
  dprint("Inserting value to stack, size %u before insert",
      lang_size(&state->value_stack));
  if(!syntetic && state->last_value == state->last_token - 1) //No operator in between
  {
    state->last_value = state->last_token;
    oper_insert(state, &space_oper, " ");// Push trough all needed operators
  }
  state->last_value = state->last_token;
  lang_push(&state->value_stack, node);
}

// Just a really nasty automata
void treeb_insert(struct treeb_state *state, const char *token,
    size_t len, enum token_type type, unsigned line)
{
  dprint("Treebuilder - insert token '%s' type %u, line %u", token, (unsigned) type, line);
  state->line = line;
  state->last_token ++;
  switch(type)
  {
    case tt_open_block:
    {
      // TODO Check we are not inside a fun block already
      struct lang_node *node = node_alloc(state);
      node->type = ln_block;
      if(state->eat_block)
      {// Terminate the last command (the one we want to add the block to)
        treeb_insert(state, "\n", 1, tt_command_end, line);
        if(state->is_fun)
          state->in_fun = true;
      }
      if(lang_size(&state->block_stack))
        langlist_append(state->eat_block ? langlist_last(
            lang_top(&state->block_stack)) : lang_top(
            &state->block_stack), node);
      state->eat_block = false;
      lang_push(&state->block_stack, node);
      break;
    }
    case tt_close_block:
    {
      oper_run_all(state);
      state->in_fun = false;
      (void) lang_pop(&state->block_stack);
      if(!lang_size(&state->block_stack))
        fatal(state, "Too many '}'s");
      break;
    }
    case tt_string_start:
      lang_push(&state->value_stack, &string_stop);
      break;
    case tt_string_data:
    case tt_string_var:
    case tt_string_var_escaped:
    {
      struct lang_node *node = node_alloc(state);
      node->type = (type == tt_string_data) ? ln_string :
        (type == tt_string_var) ? ln_variable : ln_variable_escape;
      node->data = strpool_dupe(&strpool, token, len);
      lang_push(&state->value_stack, node);
      break;
    }
    case tt_string_end:
    {
      struct lang_node *node = node_alloc(state), *part;
      node->type = ln_string_concat;
      while((part = lang_pop(&state->value_stack)) != &string_stop)
        langlist_prepend(node, part);
      value_insert(state, node, 0);
      break;
    }
    case tt_perl:
    {
      struct lang_node *node = node_alloc(state);
      node->type = ln_perl;
      node->data = strpool_dupe(&strpool, token, len);
      value_insert(state, node, 0);
      break;
    }
    case tt_command_end:
      oper_run_all(state);
      if(lang_size(&state->value_stack) > 1)
        fatal(state, "Syntax error");
      if(lang_size(&state->value_stack))//Ignore empty commands
        langlist_append(lang_top(&state->block_stack),
            lang_pop(&state->value_stack));
      break;
    case tt_open_paren:
      if(!state->in_def)
      {
        if(!lang_size(&state->value_stack) ||
            !lang_top(&state->value_stack) ||
            (lang_top(&state->value_stack)->type != ln_variable &&
            lang_top(&state->value_stack)->type != ln_perl) ||
            state->last_value < state->last_token - 1)
          // Function that does nothing - makes parenthesis work
          value_insert(state, node_alloc(state), 0);
        if(lang_top(&state->value_stack)->type != ln_perl)
          // Make the variable a function (it is called now)
          lang_top(&state->value_stack)->type = ln_function;
      }
      lang_push(&state->oper_stack, &parenthesis_stop);
      lang_push(&state->oper_stack, &comma_stop);
      lang_push(&state->value_stack, NULL);
      state->last_open_paren = state->last_token;
      break;
    case tt_close_paren:
      // First, terminate the last argument
      treeb_insert(state, "", 0, tt_comma, line);
      assert(lang_top(&state->oper_stack) == &comma_stop);
      (void) lang_pop(&state->oper_stack);
      assert(lang_top(&state->oper_stack) == &parenthesis_stop);
      (void) lang_pop(&state->oper_stack);
      assert(!lang_top(&state->value_stack));
      (void) lang_pop(&state->value_stack);
      state->in_def = state->is_map = false;
      break;
    case tt_comma:
      oper_run_until(state, &comma_stop);
      lang_push(&state->oper_stack, &comma_stop);
      if(state->in_def)
      {
        struct lang_node *node = lang_pop(&state->value_stack);//Get the single value
        if(node)
        { //Empty paretheses? TODO Make sure it really isn't something like ($a,)
          if(state->is_extract && node->type == ln_variable)
          {
            state->in_def = false;
            lang_push(&state->value_stack, node);
            goto IS_NORMAL_VAR;
          }
          assert(node->type == ln_param_multi ||
              node->type == ln_param_single);
          if(state->is_map && node->type == ln_param_multi)
            fatal(state, "map handles $ variables only");
          if(state->is_extract && node->type == ln_param_multi)
            state->in_def = false;
          assert(lang_top(&state->oper_stack) == &comma_stop);
          assert(!lang_top(&state->value_stack));
          (void) lang_pop(&state->value_stack);
          enum lang_node_type tp;
          assert((tp = lang_top(&state->value_stack)->type) ==
              ln_fun_def || tp == ln_class_def || tp == ln_map ||
              tp == ln_extract);
          langlist_append(lang_top(&state->value_stack), node);
        }
        else if(state->last_open_paren < state->last_token - 2)
          fatal(state, "Syntax error in function definition");
        lang_push(&state->value_stack, NULL);
        if(state->is_map)
          state->in_def = false;
      } else {
        IS_NORMAL_VAR:;
        struct lang_node *node = node_alloc(state);
        node->type = ln_array_concat;
        array_load(state, node);
        if(langlist_size(node) || len || state->last_open_paren <
            state->last_token - 2)//Not only empty parethesis
          langlist_append(lang_top(&state->value_stack), node);
        if(state->is_map && langlist_size(lang_top(&state->value_stack))
            > 2)
          fatal(state, "Map is for single list only");
        lang_push(&state->value_stack, NULL);
      }
      break;
    case tt_start_subscript:
      if(!lang_size(&state->value_stack) ||
          lang_top(&state->value_stack)->type != ln_variable)
        fatal(state, "What variable do you want to subscript?");
      lang_push(&state->oper_stack, &subscript_stop);
      lang_push(&state->value_stack, NULL);//Remember, where the array ends (or starts, depends on the end you begin at)
      break;
    case tt_end_subscript:
    {
      oper_run_until(state, &subscript_stop);
      struct lang_node *node = node_alloc(state);
      node->type = ln_subscript;
      array_load(state, node);
      langlist_prepend(node, lang_pop(&state->value_stack));
      lang_push(&state->value_stack, node);//Do not use value_insert, it is just node replacement
      state->last_value = state->last_token;//But if there is just space, concat it
      break;
    }
    case tt_vartype:
      if(!state->in_def)
        fatal(state, "Misplaced vartype (%s)", token);
      if(state->has_vartype)
        fatal(state, "Multiple vartypes (%s)", token);
      state->has_vartype = true;
      if(*token == '$')
        state->is_multi = false;
      else
        state->is_multi = true;
      return;
    case tt_ident:
      if(!oper_consult(NULL, token))
      {
        if(!strcmp("map", token))
        {
          struct lang_node *node = node_alloc(state);
          node->type = ln_map;
          state->is_map = state->eat_block = state->in_def = true;
          value_insert(state, node, 0);
        }
        else if(!strcmp("extract", token))
        {
          struct lang_node *node = node_alloc(state);
          node->type = ln_extract;
          state->is_extract = state->in_def = true;
          value_insert(state, node, 0);
        }
        else if(!strcmp("class", token))
        {
          state->expect_def = true;
          state->is_fun = false;
          return;
        }
        else if(!strcmp("fun", token))
        {
          state->expect_def = state->is_fun = true;
          return;
        }
        else
        {
          struct lang_node *node = node_alloc(state);
          node->data = strpool_dupe(&strpool, token, len);
          if(state->expect_def)
          {
            node->type = state->is_fun ? ln_fun_def : ln_class_def;
            state->expect_def = false;
            state->in_def = state->eat_block = true;
          }
          else if(state->has_vartype)
          {
            node->type = state->is_multi ? ln_param_multi :
              ln_param_single;
            state->has_vartype = false;
          }
          else
          {
            node->type = ln_variable;
          }
          value_insert(state, node, 0);
        }
        break;
      }
      // Leave it as operator
    case tt_operator:
    {
      struct lang_node *node = node_alloc(state);
      oper_insert(state, node, token);
      break;
    }
    default:
      assert(0);//Can not happen
  }
  assert(!state->has_vartype && !state->expect_def);
}
