/*
        Remake
        Copyright (C) 2008-2009 Michal Vaner (vorner@ucw.cz)

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        he Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "plan.h"
#include "pobject.h"

#include <common/debug.h>
#include <common/mempool_core.h>
#include <common/file.h>
#include <common/config.h>
#include <cache/files.h>
#include <lang/var.h>
#include <lang/compute.h>
#include <lang/object.h>
#include <lang/function.h>
#include <interact/coroutine.h>

#include <stddef.h>
#include <string.h>

static struct pobject_list ready, wait_deps, done, unscheduled, working,
    attached;

static struct mempool_core plan_pool_internal, *tmp_pool, dep_pool;
struct mempool_core *plan_pool = &plan_pool_internal;

void plan_init(void)
{
  mempool_core_init(plan_pool, 0, true);
  mempool_core_init(&dep_pool, 0, true);
}

static struct pobject *object_add(struct var_value *object)
{
  // TODO search if we already have this object
  // TODO search for all attached objects and add them
  struct pobject *pobject = mempool_alloc(plan_pool, sizeof *pobject);
  pobject->object = object;
  pobject_list_append(&unscheduled, pobject);
  return pobject;
}

void plan_add(struct variable *targets)
{
  for(size_t i = 0; i < targets->count; i ++)
    object_add(&targets->values[i]);
}

static void dep_add(struct pobject *owner, struct pobject *dep)
{
  dprint("%s depends on %s", owner->object->string,
      dep->object->string);
  struct dep_bind *bind = mempool_alloc(&dep_pool, sizeof *bind);
  bind->owner = owner;
  bind->dep = dep;
  dep_list_append(owner, bind);
  owner_list_append(dep, bind);
}

static void object_schedule(struct pobject *object)
{
  dprint("Scheduling %s", object->object->string);
  if(object->scheduling)
    die("Circular dependency for %s", object->object->string);
  if(object->scheduled) // Already done
  {
    dprinte("Already scheduled");
    return;
  }
  object->scheduling = true;
  if(object->object->type == vt_object)
  {
    struct variable *deps = variable_compute(object->object->object->context, "DEPS", tmp_pool);
    for(size_t i = 0; i < deps->count; i ++)
    {
      if(deps->values[i].type != vt_object && deps->values[i].type != vt_string)
        die("One of dependency of %s is not an object or string", object->object->string);
      struct pobject *dep = object_add(&deps->values[i]);
      dep_add(object, dep);
      object_schedule(dep);
    }
  }
  else
    dprint("%s is just an input file, no deps", object->object->string);
  object->scheduled = true;
  object->scheduling = false;
}

static void deps_compute(void)
{
  dprinte("Computing dependencies");
  struct pobject *current;
  while(current = pobject_list_take_first(&unscheduled))
  {
    object_schedule(current);
    pobject_list_append(current->dep_head ? &wait_deps : &ready,
        current);
  }
}

static void dep_done(struct dep_bind *dep, bool *changed)
{
  if(*changed)
    dep->owner->changed_deps = true;
  dep_list_remove(dep->owner, dep);
  if(!dep_list_first(dep->owner)) // No deps
  {
    pobject_list_remove(&wait_deps, dep->owner);
    pobject_list_append(&ready, dep->owner);
  }
}

static void task_done(struct pobject *object, struct object *master)
{
  if(master && object->object->object != master)
    return;
  object->done = true;
  pobject_list_remove(master ? &attached : &working, object);
  pobject_list_append(&done, object);
  bool changed = object->changed_deps;
  owner_list_walk(object, &changed, (owner_list_walk_func) dep_done);
  owner_list_drop(object);
  if(!master)
    pobject_list_walk(&attached, object->object->object, (pobject_list_walk_func) task_done);
}

static void attached_execute(struct pobject *pobject, struct object *object)
{
  if(pobject->object->object == object)
  {
    dprint("Attached object %s", pobject->object->string);
    pobject_list_remove(&ready, pobject);
    pobject_list_append(&attached, pobject);
  }
}

struct object_task
{
  struct pobject *pobject;
  bool changed_deps:1;
};

static void method_run(struct object *object, const char *obj_name, const char *method_name,
    struct mempool_core *tmp_pool)
{
  struct variable *method = variable_compute(object->context, method_name, tmp_pool);
  if(!method->count)
  {
    dprint("Method %s of object %s does not exist, skipping", method_name, obj_name);
  }
  else
  {
    if(method->count != 1)
      die("Method %s of object %s must have exactly 1 value", method_name, obj_name);
    if(method->values[0].type != vt_method)
      die("%s in object %s is not a method", method_name, obj_name);
    function_apply(method->values[0].function, 0, NULL, object->context, object->context, tmp_pool, true,
        method->values[0].function->func, true);
  }
}

static void object_execute_internal(struct object_task *task_remote)
{
  dprinte("Object execute internal");
  // Get it to local stack so remote can be destroyed in context-switch
  struct object_task task = *task_remote;
  struct object *object = task.pobject->object->object;
  struct mempool_core obj_tmp_pool;
  mempool_core_init(&obj_tmp_pool, 0, true);
  // Run both methods (default and action, if needed)
  bool obey_deps = true;
  if(config_methods.next)
  {
    struct variable *actions = variable_compute(object->context, "ACTIONS", &obj_tmp_pool);
    obey_deps = false;
    struct strlist *method = &config_methods;
    while(method = method->next)
      for(size_t i = 0; i < actions->count; i ++)
      {
        assert(actions->values[i].type == vt_string);
        if(!strcmp(actions->values[i].string, method->string))
        {
          obey_deps = true;
          break;
        }
      }
  }
  if(task.changed_deps && obey_deps)
    method_run(object, task.pobject->object->string, "DEFAULT", &obj_tmp_pool);
  struct strlist *method = &config_methods;
  while(method = method->next)
    method_run(object, task.pobject->object->string, method->string, &obj_tmp_pool);
  task_done(task.pobject, NULL);
  // Prepare for next time
  if(*task.pobject->object->string)
    cache_file_mark(task.pobject->object->string);  // Mark in the cache for future
  struct variable *new_deps = variable_compute(object->context, "DEPS", &obj_tmp_pool);
  // Mark the new deps in the cache
  for(size_t i = 0; i < new_deps->count; i ++)
  {
    if(new_deps->values[i].type != vt_string && new_deps->values[i].type != vt_object)
      die("One of the dependencies of %s is not an object or string", task.pobject->object->string);
    if(*new_deps->values[i].string)
      cache_file_mark(new_deps->values[i].string);
  }
  mempool_core_clean(&obj_tmp_pool);
}

static void object_execute(struct pobject *object, bool chaged_deps)
{
  pobject_list_walk(&ready, object->object->object, (pobject_list_walk_func) attached_execute);
  coroutine_start((coroutine_func) object_execute_internal, &(struct object_task) {
      .pobject = object,
      .changed_deps = chaged_deps
    });
}

static void plan_execute(void)
{
  struct pobject *current;
  while((current = pobject_list_take_first(&ready)) || pobject_list_first(&working))
  {
    if(current)
    {
      dprint("Execute object %s", current->object->string);
      pobject_list_append(&working, current);
      if(current->object->type == vt_string)
      {
        if(!file_exists(current->object->string))
          die("Needed file %s has no rule to be made and does not exist" , current->object->string);
        current->changed_deps = cached_file_changed(current->object->string);
        task_done(current, NULL);
      }
      else
      {
        if((*current->object->string && (!file_exists(current->object->string) ||
            cached_file_changed(current->object->string))) || !current->object->object->cache_loaded)
          current->changed_deps = true;
        bool deps = false;
        if(config_run_all)
        {
          dprinte("Requested to run everything");
          deps = true;
        }
        else if(current->changed_deps)
        {
          dprinte("Needs to be re-run");
          deps = true;
        }
        if(deps || config_methods.next)
          object_execute(current, deps);
        else
        {
          dprinte("No deps changed, not re-running");
          task_done(current, NULL);
        }
      }
    }
    // Pick up any dead children, their outputs, etc
    coroutine_scheduler(!current);
  }
  assert(!pobject_list_first(&wait_deps));
  assert(!pobject_list_first(&attached));
}

void plan_run(struct mempool_core *tmp_pool_)
{
  tmp_pool = tmp_pool_;
  deps_compute();
  mempool_core_clean(tmp_pool);
  plan_execute();
  // The tmp_pool is not used here, we use separate for
  // each object
  mempool_core_clean(&dep_pool);
}

void plan_clean(void)
{
  mempool_core_clean(plan_pool);
}
